/*****************************************************************************
 *									     *
 *  THIS IS A C DRIVER FOR THE DGETRF FORTRAN SUBROUTINE.		     *
 *  C MATRIX INDICES ARE TRANSPOSED AND THEIR RANGES ARE 1:N		     *
 *  INSTEAD OF 0:N-1, IN ORDER TO MATCH THE FORTRAN MATRICES CONVENTIONS.    *
 *  Giuseppe Ciaccio 24 May 1998					     *
 *									     *
 *****************************************************************************/

	/******************************************************
	 *  Parallel version:				      *
	 *  BLK_SIZE is both block size and width of each     *
	 *  column strip by which the matrix is distributed   *
	 *  among processors.				      *
	 ******************************************************/

#define PARALLEL
#undef MEASURE_LOOP
#define BACK_SUB
#define NMAX 2100
#define LDA NMAX
#define REAL double
#define ZERO 0.0e0
#define ONE 1.0e0
#undef TIME
#undef PERF_SHORT

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <linux/gamma/libgamma.h>
#include "measure.h"

void print_time (int row);
void dmxpy (int n1, REAL y[], int n2, int ldm, REAL x[], REAL m[]);
REAL epslon (REAL x);
extern int dgetrs_();
extern int daxpy_();


#ifdef PARALLEL

#define NPROC 2
#define BLK_SIZE 64
#define N_PIV_BUFS 3
#define gamma_SEND(a,b,c) gamma_send_fast(a,b,c)
#define MAT_PORT 1
#define PIV_PORT 2
#define ACK_PORT 3
#define B_INIT_PORT 4
#define BACK_PORT_1 5
#define BACK_PORT_2 6
#define min(a,b) (((a)<(b))?(a):(b))

REAL *bi;
REAL *piv_buf;
int piv_indx_row,piv_indx_col;
	/*
	 *  NOTE:
	 *  indices are reversed to match FORTRAN matrix layout!
	 */
int piv_buf_tail;
int b_indx;
int whoami;
int mat_port[NPROC];

extern int pdgetrf_();
extern int dlaswp_();
extern int dtrsm_();
void pmatgen (REAL a[], int lda, int n, REAL b[], REAL *norma, REAL a_bis[]);
void pdgetrs(int n, REAL *a, int lda, int *ipvt, REAL *b, int blk_size);

static inline int owner(int whoami, int k) {
if (whoami==0) whoami = NPROC;
return ((whoami-1) == ((k-1)/BLK_SIZE) % NPROC) ? 1 : 0;
}

static inline int virt_to_phys(int whoami, int k) {
int a,b;
if (whoami==0) whoami = NPROC;
a = (k%(BLK_SIZE*NPROC))/BLK_SIZE;
b = ((whoami<=a) ? BLK_SIZE : ((whoami == a+1) ? k%BLK_SIZE : 0));
return b+BLK_SIZE*(k/(BLK_SIZE*NPROC));
}

static void piv_handler(void) {
        gamma_signal(PIV_PORT);
	if (piv_indx_col==BLK_SIZE) {
		/*  A new block column has been delivered.  */
		piv_indx_row += BLK_SIZE;
	        piv_indx_col = 1;
		if (piv_buf_tail==N_PIV_BUFS)
			piv_buf_tail=1;
		else
			piv_buf_tail++;
	}
	else
	        piv_indx_col++;
	/*
	 *  "piv_buf_tail", "piv_indx_row" and "piv_indx_col"
	 *  are FORTRAN matrix indices.
	 *  FORTRAN indices range from 1 to N, whereas
	 *  C indices range from 0 to N-1.
	 *  Therefore they must be decremented by 1 when used as subscripts
	 *  in C expressions.
	 */
        gamma_attach_buffer(PIV_PORT,&piv_buf[
		BLK_SIZE*LDA*(piv_buf_tail-1)
		+LDA*(piv_indx_col-1)
		+piv_indx_row-1]);
}

static void back_2_handler(void) {
        gamma_signal(BACK_PORT_2);
        b_indx -= BLK_SIZE;
        gamma_attach_buffer(BACK_PORT_2,&bi[b_indx-1]);
}

static void generic_handler(void) {
        gamma_signal(gamma_active_port);
}

static void err_handler(void) {
        gamma_sigerr(gamma_active_port);
}

#else	/*  not parallel  */

#define NPROC 1
void matgen (REAL a[], int lda, int n, REAL b[], REAL *norma);
extern int dgetrf_();

#endif


#ifdef MEASURE_LOOP

#ifdef PARALLEL
#define START_LOOP (BLK_SIZE*NPROC-1)
#define STOP_LOOP 2000
#define INCR_LOOP (BLK_SIZE*NPROC)
/*
#define START_LOOP 1999
#define STOP_LOOP 3999
#define INCR_LOOP 1000
*/
#else
#define START_LOOP 16
#define STOP_LOOP 255
#define INCR_LOOP 1
#endif

#else	/*  no measure loop  */

#define N 300	/*  Must be less than NMAX  */

#endif


int main (int argc, char **argv)
{
#ifdef PARALLEL
	static REAL *pivbuf;
#endif
	static REAL *a;
	static REAL *b;
	static REAL *x;

	static int ipvt_vector[NMAX+32/sizeof(int)];
	static int *ipvt=ipvt_vector;

	int j,info,lda=LDA,n;

        REAL ops,norma;
#ifndef MEASURE_LOOP
#ifndef PERF_SHORT
        REAL normx,resid,residn,eps,epsn,x1,x2;
#endif
#endif

#ifdef PARALLEL
	int	nproc=NPROC,
		blk_size=BLK_SIZE,
		p_row_port = PIV_PORT,
		n_piv_bufs = N_PIV_BUFS;
#else
	int one=1;
	char trans[4]="Tran";
#endif

        time_586 t1,t2,t3;

	double
		record_lu_lda[NPROVE],
		record_solv_lda[NPROVE],
		record_tot_lda[NPROVE];

	double
		m_lu_lda, med_lu_lda, M_lu_lda, var_lu_lda,
		m_solv_lda, med_solv_lda, M_solv_lda, var_solv_lda, 
		m_tot_lda, med_tot_lda, M_tot_lda, var_tot_lda;


#ifdef PARALLEL
/*	a = (REAL *)malloc(NMAX*LDA*sizeof(REAL)/NPROC+32);*/
	a = (REAL *)malloc(
((((NMAX+BLK_SIZE)/BLK_SIZE)+NPROC)/NPROC)*BLK_SIZE*LDA*sizeof(REAL)+32
	);
	/*
	 *  I don't know why twice the size is necessary here...
	 */
	pivbuf = (REAL *)malloc(2*N_PIV_BUFS*BLK_SIZE*LDA*sizeof(REAL)+32);
#else
	a = (REAL *)malloc(NMAX*LDA*sizeof(REAL)+32);
#endif
	b = (REAL *)malloc(NMAX*sizeof(REAL)+32);
	x = (REAL *)malloc(NMAX*sizeof(REAL)+32);

	/*  Aligning arrays on cache lines boundaries...  */
	a = (REAL *)(((unsigned long)((char *)a+31))&~31);
	b = (REAL *)(((unsigned long)((char *)b+31))&~31);
	x = (REAL *)(((unsigned long)((char *)x+31))&~31);
#ifdef PARALLEL
	pivbuf = (REAL *)(((unsigned long)((char *)pivbuf+31))&~31);
#endif
	ipvt = (int *)(((unsigned long)((char *)ipvt+31))&~31);

#ifdef PARALLEL
        gamma_init(NPROC,argc,argv);
        whoami = gamma_my_node();
        for (j=0;j<NPROC;j++) {
                mat_port[j] = j+10;
                gamma_setport(mat_port[j],
                        j,gamma_my_par_pid(),MAT_PORT,
                        NULL,NULL,NULL,0
                );
        }
        if (whoami == 0)
                gamma_setport(ACK_PORT,
                        BROADCAST,gamma_my_par_pid(),ACK_PORT,
                        generic_handler,err_handler,NULL,0
                );
        else
                gamma_setport(ACK_PORT,
                        0,gamma_my_par_pid(),ACK_PORT,
                        generic_handler,err_handler,NULL,0
                );
        piv_buf = pivbuf;
        bi = b;

	/*
	 *  I don't know why twice the size is necessary here...
	 */
        gamma_mlock(pivbuf,2*N_PIV_BUFS*BLK_SIZE*LDA*sizeof(REAL *));

        gamma_mlock(&piv_buf,sizeof(REAL *));
        gamma_mlock(&piv_indx_row,sizeof(int));
        gamma_mlock(&piv_indx_col,sizeof(int));
        gamma_mlock(&piv_buf_tail,sizeof(int));
        gamma_mlock(&bi,sizeof(REAL *));
        gamma_mlock(&b_indx,sizeof(int));
#endif


#ifdef MEASURE_LOOP
for (n=START_LOOP;n<=STOP_LOOP;n+=INCR_LOOP) {
	/*
	 *  Skip multiples of 64 from 256 on,
	 *  which appear to lead to great instability.
	 */
/*	if ((n>=256)&&(!(n%64))) continue;*/
#else
        n = N;
#endif

#ifdef BACK_SUB
        ops =
(2.0e0*((double)n*(double)n*(double)n))/3.0 + 2.0*((double)n*(double)n);
#else
        ops =
(2.0e0*((double)n*(double)n*(double)n))/3.0 + 0.5*((double)n*(double)n);
#endif

#ifdef PARALLEL
        gamma_setport(MAT_PORT,
                DONT_MIND,DONT_MIND,DONT_MIND,
                generic_handler,err_handler,NULL,n*sizeof(REAL)
        );
        gamma_setport(PIV_PORT,
                BROADCAST,gamma_my_par_pid(),PIV_PORT,
                piv_handler,err_handler,NULL,n*sizeof(REAL)
        );
	gamma_setport(B_INIT_PORT,
		BROADCAST,gamma_my_par_pid(),B_INIT_PORT,
		generic_handler,err_handler,b,n*sizeof(REAL)
	);
	gamma_setport(BACK_PORT_1,
		BROADCAST,gamma_my_par_pid(),BACK_PORT_1,
		generic_handler,err_handler,NULL,n*sizeof(REAL)
	);
	gamma_setport(BACK_PORT_2,
		BROADCAST,gamma_my_par_pid(),BACK_PORT_2,
		back_2_handler,err_handler,NULL,BLK_SIZE*sizeof(REAL)
	);
        gamma_sync();
#endif

#ifndef MEASURE_LOOP
#ifndef PERF_SHORT
#ifdef PARALLEL
	{
	static REAL *a_bis;
        pmatgen(a,lda,n,b,&norma,NULL);
	pdgetrf_(&n,&n,a,&lda,ipvt,&info,
		&nproc,&blk_size,&p_row_port,&piv_indx_row,&piv_indx_col,
		pivbuf,&piv_buf_tail,&n_piv_bufs);
	pdgetrs(n,a,lda,ipvt,b,BLK_SIZE);
#else
        matgen(a,lda,n,b,&norma);
	dgetrf_(&n,&n,a,&lda,ipvt,&info);
	dgetrs_(trans,&n,&one,a,&lda,ipvt,b,&n,&info);
#endif

/*	compute a residual to verify results.  */ 
        for (j = 0; j < n; j++) x[j] = b[j];

#ifdef PARALLEL
	if (whoami==0)
		a_bis = (REAL *)malloc(N*LDA*sizeof(REAL));
	else
		a_bis = NULL;
        pmatgen(a,lda,n,b,&norma,a_bis);
#else
        matgen(a,lda,n,b,&norma);
#endif
        for (j = 0; j < n; j++) {
                b[j] = -b[j];
        }
#ifdef PARALLEL
	if (whoami == 0) {
	dmxpy(n,b,n,lda,x,a_bis);
	free((void *)a_bis);
#else
        dmxpy(n,b,n,lda,x,a);
#endif
        resid = 0.0;
        normx = 0.0;
        for (j = 0; j < n; j++) {
                resid = (resid > fabs((double)b[j])) 
                        ? resid : fabs((double)b[j]);
                normx = (normx > fabs((double)x[j])) 
                        ? normx : fabs((double)x[j]);
        }
        eps = epslon(ONE);
        residn = resid/( n*norma*normx*eps );
        epsn = eps;
        x1 = x[0] - 1;
        x2 = x[n-1] - 1;

        printf("norm resid      resid           machep");
        printf("         x[0]-1          x[n-1]-1\n");
        printf("%6.1f %17.8e%17.8e%17.8e%17.8e\n\n",
               (double)residn, (double)resid, (double)epsn, 
               (double)x1, (double)x2);
        printf("Times are reported for matrices of order    %5d\n",n);

#ifdef PARALLEL
        }
        }
#endif
#else	/*  perf_short  */
#ifdef PARALLEL
	gamma_sync();
	printf("LINPACK %dx%d on %d nodes: node %d starting...\n",
		n,n,NPROC,whoami);
	fflush(stdout);
#endif
#endif	/*  perf_short  */
#endif	/*  no measure loop  */


/************************************************************************
 *                              Execute NPROVE passes with lda          *
 ************************************************************************/
      
        for (j=0 ; j<NPROVE ; j++) {
#ifdef PARALLEL
	        pmatgen(a,lda,n,b,&norma,NULL);
		gamma_time(t1);
		pdgetrf_(&n,&n,a,&lda,ipvt,&info,
			&nproc,&blk_size,&p_row_port,
			&piv_indx_row,&piv_indx_col,
			pivbuf,&piv_buf_tail,&n_piv_bufs);
#else
	        matgen(a,lda,n,b,&norma);
		gamma_time(t1);
		dgetrf_(&n,&n,a,&lda,ipvt,&info);
#endif
		gamma_time(t2);
#ifdef BACK_SUB
#ifdef PARALLEL
		pdgetrs(n,a,lda,ipvt,b,BLK_SIZE);
#else
		dgetrs_(trans,&n,&one,a,&lda,ipvt,b,&n,&info);
#endif
#endif
		gamma_time(t3);
		record_lu_lda[j] = gamma_time_diff(t2,t1);
		record_solv_lda[j] = gamma_time_diff(t3,t2);
		record_tot_lda[j] = record_lu_lda[j]+record_solv_lda[j];
        }

/************************************************************************
 *           Compute measure results                                    *
 ************************************************************************/

	mean_variance(record_lu_lda,&m_lu_lda,&M_lu_lda,&med_lu_lda,
		&var_lu_lda);
	mean_variance(record_solv_lda,&m_solv_lda,&M_solv_lda,&med_solv_lda,
		&var_solv_lda);
	mean_variance(record_tot_lda,&m_tot_lda,&M_tot_lda,&med_tot_lda,
		&var_tot_lda);

/************************************************************************
 *           Print measure results                                      *
 ************************************************************************/

#ifdef PARALLEL
	if (whoami == 0) {
#endif

#ifdef MEASURE_LOOP
#ifdef TIME
	printf("%d %d %g\n",NPROC,n,med_tot_lda/1e6);
#else
	printf("%d %d %g\n",NPROC,n,ops/med_tot_lda);
#endif
	fflush(stdout);
#else
	printf("\n\n");
	printf("PERFORMANCE (secs) (best < average < worse):\n\n");
	printf("LU factoring\t\t%8f\t< %8f\t< %8f\n",
		m_lu_lda/1e6,med_lu_lda/1e6,M_lu_lda/1e6);
	printf("Backward substitution\t%8f\t< %8f\t< %8f\n",
		m_solv_lda/1e6,med_solv_lda/1e6,M_solv_lda/1e6);
	printf("Total\t\t\t%8f\t< %8f\t< %8f\n",
		m_tot_lda/1e6,med_tot_lda/1e6,M_tot_lda/1e6);
	printf("MFLOPS\t\t\t%6f\t> %6f\t> %6f\n",
		ops/m_tot_lda,ops/med_tot_lda,ops/M_tot_lda);
#endif
#ifdef PARALLEL
	}
#endif
#ifdef MEASURE_LOOP
}
#endif

#ifdef PARALLEL
        gamma_munlockall();
	gamma_exit();
#endif

	return 0;
}
     
/*----------------------*/ 

#ifdef PARALLEL

void pmatgen (REAL a[], int lda, int n, REAL b[], REAL *norma, REAL a_bis[])
/*
	PARALLEL VERSION.
	Output:
	Matrix "a" is scattered among nodes according to the owner() function,
	which must ensure that the last matrix strip is assigned
	to node zero (master node).
	On node zero, matrix "a_bis" contains a duplicate of the whole
	matrix a.
	Vector "b" is replicated at each node.
	Scalar "norma" is meaningful only on node zero.
	MATRIX INDICES REVERSED TO MATCH FORTRAN MATRIX LAYOUT!!
	References to FORTRAN a(i,j) are written a[lda*(j-1)+(i-1)].
*/
{
	REAL aa[LDA];
        int w,init, i, j;

	if (whoami == 0) {
		/*  Master generates and distributes matrix rows.  */
		/*  Master owns last matrix strip.  */
		init = 1325;
		*norma = 0.0;
		for (i = 1; i <= n; i++) b[i-1] = 0.0;
		for (i = 1; i <= n; i++) {
			/*  Compute i-th column.  */
			for (j = 1; j <= n; j++) {
				init = 3125*init % 65536;
				aa[j-1] = (init - 32768.0)/16384.0;
				if (a_bis!=NULL)
					a_bis[lda*(i-1)+(j-1)] = aa[j-1];
				*norma = (aa[j-1] > *norma) ? aa[j-1] : *norma;
				b[i-1] = b[i-1] + aa[j-1];
			}

			/*  Put i-th column to the correct destination.  */

			/*  Find the owner  */
			for (j=0;j<NPROC;j++) {
				if (owner(j,i)) {
					w=j;
					break;
				}
				w=-1;
			}

			if (w < 0)
			printf("pmatgen(): ERROR: no owner for col %d\n",i);

			gamma_SEND(ACK_PORT,NULL,0);

			if (w != whoami) {
				/*  Owner of col. "i" is remote.  */
				gamma_wait(ACK_PORT,1);
				gamma_mlock(aa,n*sizeof(REAL));
				gamma_SEND(mat_port[w],aa,
					n*sizeof(REAL));
				gamma_munlock(aa,n*sizeof(REAL));
			}
			else {
				/*  Owner of col. "i" is master node.  */
				for (j = 1; j <= n; j++)
a[lda*(virt_to_phys(whoami,i)-1)+(j-1)] = aa[j-1];
			}
		}
	}

	else {
		/*  Slaves receive columns from master  */
		for (i=1;i<=n;i++) {
			gamma_wait(ACK_PORT,1);
			if (owner(whoami,i)) {
				gamma_attach_buffer(MAT_PORT,
					&a[lda*(virt_to_phys(whoami,i)-1)]
				);
				gamma_mlock(
					&a[lda*(virt_to_phys(whoami,i)-1)],
					n*sizeof(REAL)
				);
				gamma_SEND(ACK_PORT,NULL,0);
				gamma_wait(MAT_PORT,1);
				gamma_munlock(
					&a[lda*(virt_to_phys(whoami,i)-1)],
					n*sizeof(REAL)
				);
			}
		}
	}

	/*  Master bcasts vector b among slaves.  */
	if (!whoami) {
		gamma_mlock(b,n*sizeof(REAL));
		gamma_sync();
		gamma_SEND(B_INIT_PORT,b,n*sizeof(REAL));
		gamma_munlock(b,n*sizeof(REAL));
	}
	else {
		gamma_mlock(b,n*sizeof(REAL));
		gamma_sync();
		gamma_wait(B_INIT_PORT,1);
		gamma_munlock(b,n*sizeof(REAL));
	}

	return;
}

void pdgetrs(int n, REAL *a, int lda, int *ipvt, REAL *b, int blk_size) {
	/*  Distributed Backward Substitution.  */
	/*  References to FORTRAN a(i,j) are written a[lda*(j-1)+(i-1)].  */
	REAL temp;
	int i,ii,jb=blk_size,k,one=1,minus_one=-1;
	gamma_mlock(b,n*sizeof(REAL));
	gamma_attach_buffer(BACK_PORT_1,b);
	gamma_sync();
	for (ii=1; ii<=n; ii+=blk_size) {
		jb = min(blk_size,n-ii+1);
		if (owner(whoami,ii)) {
			for (i=ii; i<ii+jb; i++) {
				temp = b[i-1];
				for (k=1; k<i; k++)
temp -= a[lda*(virt_to_phys(whoami,i)-1)+(k-1)]*b[k-1];
b[i-1] = temp/a[lda*(virt_to_phys(whoami,i)-1)+(i-1)];
			}
			gamma_attach_buffer(BACK_PORT_1,&b[ii+jb-1]);
			gamma_SEND(BACK_PORT_1,&b[ii-1],jb*sizeof(REAL));
		}
		else
			gamma_wait(BACK_PORT_1,1);
	}
	gamma_attach_buffer(BACK_PORT_1,&b[n-jb+1-1]);
	b_indx = n-jb+1-blk_size;
	gamma_attach_buffer(BACK_PORT_2,&b[b_indx-1]);
	gamma_sync();
	if (owner(whoami,n)) {
		for (i=n; i>=n-jb+1; i--) {
			temp = b[i-1];
			for (k=i+1; k<=n; k++)
temp -= a[lda*(virt_to_phys(whoami,i)-1)+(k-1)]*b[k-1];
			b[i-1] = temp;
		}
		gamma_SEND(BACK_PORT_1,&b[n-jb+1-1],jb*sizeof(REAL));
	}
	else
		gamma_wait(BACK_PORT_1,1);
	for (ii=n-jb; ii>1; ii-=blk_size) {
		if (owner(whoami,ii)) {
			for (i=ii; i>=ii-blk_size+1; i--) {
				temp = b[i-1];
				for (k=i+1; k<=n; k++)
temp -= a[lda*(virt_to_phys(whoami,i)-1)+(k-1)]*b[k-1];
				b[i-1] = temp;
			}
			/*  Handler-like operations.  */
			b_indx -= BLK_SIZE;
			gamma_attach_buffer(BACK_PORT_2,&b[b_indx-1]);
			gamma_SEND(BACK_PORT_2,&b[ii-blk_size+1-1],
				blk_size*sizeof(REAL));
		}
		else
			gamma_wait(BACK_PORT_2,1);
	}
	dlaswp_(&one,b,&lda,&one,&n,ipvt,&minus_one);
	/*  At this point each node has a copy of the whole solution.  */
	gamma_munlock(b,n*sizeof(REAL));
	return;
}

#else	/*  not parallel  */

void matgen (REAL a[], int lda, int n, REAL b[], REAL *norma)
{
        int init, i, j;

        init = 1325;
        *norma = 0.0;
        for (j = 0; j < n; j++) {
                for (i = 0; i < n; i++) {
                        init = 3125*init % 65536;
                        a[lda*i+j] = (init - 32768.0)/16384.0;                        
                        *norma = (a[lda*i+j] > *norma) ? a[lda*i+j] : *norma;
                }
        }
        for (i = 0; i < n; i++) {
          b[i] = 0.0;
        }
        for (j = 0; j < n; j++) {
                for (i = 0; i < n; i++) {
                        b[i] = b[i] + a[lda*i+j];
                }
        }
        return;
}

#endif

/*----------------------*/ 

REAL epslon (REAL x)
/*
     estimate unit roundoff in quantities of size x.
*/
{
        REAL a,b,c,eps;
/*
     this program should function properly on all systems
     satisfying the following two assumptions,
        1.  the base used in representing dfloating point
            numbers is not a power of three.
        2.  the quantity  a  in statement 10 is represented to 
            the accuracy used in dfloating point variables
            that are stored in memory.
     the statement number 10 and the go to 10 are intended to
     force optimizing compilers to generate code satisfying 
     assumption 2.
     under these assumptions, it should be true that,
            a  is not exactly equal to four-thirds,
            b  has a zero for its last bit or digit,
            c  is not exactly equal to one,
            eps  measures the separation of 1.0 from
                 the next larger dfloating point number.
     the developers of eispack would appreciate being informed
     about any systems where these assumptions do not hold.

     *****************************************************************
     this routine is one of the auxiliary routines used by eispack iii
     to avoid machine dependencies.
     *****************************************************************

     this version dated 4/6/83.
*/

        a = 4.0e0/3.0e0;
        eps = ZERO;
        while (eps == ZERO) {
                b = a - ONE;
                c = b + b + b;
                eps = fabs((double)(c-ONE));
        }
        return(eps*fabs((double)x));
}
 
/*----------------------*/ 

void dmxpy (int n1, REAL y[], int n2, int ldm, REAL x[], REAL m[])
/*
   purpose:
     multiply matrix m times vector x and add the result to vector y.

   parameters:

     n1 integer, number of elements in vector y, and number of rows in
         matrix m

     y double [n1], vector of length n1 to which is added 
         the product m*x

     n2 integer, number of elements in vector x, and number of columns
         in matrix m

     ldm integer, leading dimension of array m

     x double [n2], vector of length n2

     m double [n2][ldm], matrix of n1 rows and n2 columns

 ----------------------------------------------------------------------
*/
{
        int j,i,jmin;
        /* cleanup odd vector */

        j = n2 % 2;
        if (j >= 1) {
                j = j - 1;
                for (i = 0; i < n1; i++) 
                        y[i] = (y[i]) + x[j]*m[ldm*i+j];
        } 

        /* cleanup odd group of two vectors */

        j = n2 % 4;
        if (j >= 2) {
                j = j - 1;
                for (i = 0; i < n1; i++)
                        y[i] = ( (y[i])
                               + x[j-1]*m[ldm*i+(j-1)]) + x[j]*m[ldm*i+j];
        } 

        /* cleanup odd group of four vectors */

        j = n2 % 8;
        if (j >= 4) {
                j = j - 1;
                for (i = 0; i < n1; i++)
                        y[i] = ((( (y[i])
                               + x[j-3]*m[ldm*i+(j-3)]) 
                               + x[j-2]*m[ldm*i+(j-2)])
                               + x[j-1]*m[ldm*i+(j-1)]) + x[j]*m[ldm*i+j];
        } 

        /* cleanup odd group of eight vectors */

        j = n2 % 16;
        if (j >= 8) {
                j = j - 1;
                for (i = 0; i < n1; i++)
                        y[i] = ((((((( (y[i])
                               + x[j-7]*m[ldm*i+(j-7)]) + x[j-6]*m[ldm*i+(j-6)])
                               + x[j-5]*m[ldm*i+(j-5)]) + x[j-4]*m[ldm*i+(j-4)])
                               + x[j-3]*m[ldm*i+(j-3)]) + x[j-2]*m[ldm*i+(j-2)])
                               + x[j-1]*m[ldm*i+(j-1)]) + x[j]  *m[ldm*i+j];
        } 
        
        /* main loop - groups of sixteen vectors */

        jmin = (n2%16)+16;
        for (j = jmin-1; j < n2; j = j + 16) {
                for (i = 0; i < n1; i++) 
                        y[i] = ((((((((((((((( (y[i])
                                + x[j-15]*m[ldm*i+(j-15)]) 
                                + x[j-14]*m[ldm*i+(j-14)])
                                + x[j-13]*m[ldm*i+(j-13)]) 
                                + x[j-12]*m[ldm*i+(j-12)])
                                + x[j-11]*m[ldm*i+(j-11)]) 
                                + x[j-10]*m[ldm*i+(j-10)])
                                + x[j- 9]*m[ldm*i+(j- 9)]) 
                                + x[j- 8]*m[ldm*i+(j- 8)])
                                + x[j- 7]*m[ldm*i+(j- 7)]) 
                                + x[j- 6]*m[ldm*i+(j- 6)])
                                + x[j- 5]*m[ldm*i+(j- 5)]) 
                                + x[j- 4]*m[ldm*i+(j- 4)])
                                + x[j- 3]*m[ldm*i+(j- 3)]) 
                                + x[j- 2]*m[ldm*i+(j- 2)])
                                + x[j- 1]*m[ldm*i+(j- 1)]) 
                                + x[j]   *m[ldm*i+j];
        }
        return;
}
