#undef OLDSTYLE
#define SYNC_SUBSETS 2	/*  Barrier sync algorithm, half-duplex case.  */

/* luigi: prima o poi togliere?
 * Use software broadcast in initialization routines,
 * if support for multiple nics is enabled. Used only for testing.
 */
#ifdef MORE_NICS_SAMENODE_ROUTING
#define SYNC_SAFE_BCAST_SW
#else
#undef SYNC_SAFE_BCAST_SW
#endif

#include "libgamma.h"
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/user.h>	/*  "#define PAGE_SIZE ..."  */
#include <signal.h>

#include "gamma_trap.h"

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)

#ifdef USE_FLOWCTL_IN_SYNC
#define gamma_SEND(a,b,c) gamma_send_flowctl(a,b,c)
#else
#define gamma_SEND(a,b,c) gamma_send(a,b,c)
#endif
#define gamma_SEND_FLOWCTL(a,b,c) gamma_send_flowctl(a,b,c)

inline int gamma_poll(void);
inline int gamma_poll_nowait(void);
static inline int gamma_mlock_lib(void *,unsigned long,int);
static int gamma_register(u_int32_t);
static int gamma_register_nospawn(u_int32_t);
static int gamma_sync_safe(void);
static int gamma_abort_(u_int32_t);

/*
 *  The following tables and data structures are in user-space,
 *  to allow receiver handlers to change their content
 *  without performing system calls.
 */

/*
 *  The following tables have one entry per communication port.
 *  Initialized by gamma_init() and set by gamma_connect_port().
 *  Each generic entry e[i] is relative to port no. "i".
 *  The meaning of each entry is as follows:
 */
typedef struct {
	unsigned long in_use;	/* 1 if port is set for input		*/
	unsigned long sem_idx;	/* # of notif/err semaphore		*/
	void (*rec_handler)();	/* rec handler; NULL if no handler	*/
	unsigned long handler_has_run;
		/* 1 if handler has run, 0 else		*/
} port_desc;

/*  ptrs to recv handler  */
port_desc gamma_port[MAX_PORT_NUMBER+1];

/*  Mutex on notif semaphores and handler execution  */
struct { volatile int s; } gamma_spinlock;

/*
 *  Table of synchronization semaphores.
 *
 *  GAMMA semaphores are useful to synchronize recv handlers
 *  with the main thread in a safe way.
 *
 *  The correct cooperation between recv handlers and main thread
 *  can be accomplished in many different ways, however GAMMA semaphores
 *  will provide safe and easy-to-use support to such cooperation.
 *
 *  GAMMA semaphores are not classical semaphores:
 *  there is no queue for waiting processes, and the waiting is a
 *  busy one.
 *
 *  During the waiting, the NIC is inspected for message arrivals.
 *
 *  Initialized by gamma_init() and gamma_connect_port().
 *  Incremented by gamma_signal(), decremented by gamma_wait().
 *
 *  Each semaphore requires two variables for the current value, plus one
 *  variable for the current error count.  Hence the three tables.
 *  The current value of semaphore "sem" is given by
 *  sync_sem1[sem] - sync_sem2[sem].
 *  Look at gamma_signal() and gamma_wait() to understand why.
 */
volatile unsigned long sync_sem1[MAX_SYNC_SEM_NUMBER+1];
volatile unsigned long sync_sem2[MAX_SYNC_SEM_NUMBER+1];

volatile unsigned long pre_sync_sem[MAX_SYNC_SEM_NUMBER+1];
	/*  To be used only if GAMMA uses the DAMNICK device.  */

volatile unsigned long sync_err[MAX_SYNC_SEM_NUMBER+1];

/*
 *  Key of the virtual GAMMA on which this process runs
 *  (the process group which this process belongs to).
 *  Initialized by gamma_init() or gamma_init_nospawn(), reset by gamma_exit().
 */
static u_int32_t my_rnd_key = 0;

/*
 *  Id. number of the virtual GAMMA on which this process runs
 *  (the process group which this process belongs to), and associated key.
 *  Initialized by gamma_init_stub(), reset by gamma_exit() and gamma_abort();
 */
static int my_par_pid = 0;

/*
 *  The "rank" of this process, relative to this virtual GAMMA.
 *  Initialized by gamma_init_stub(), reset by gamma_exit().
 */
static int my_instnum = 0;

/*
 *  The "processor number" of this process, relative to the local node.
 *  If the local node runs, say, four process instances of a given job, then
 *  they have "procnum" resp. of 0,1,2, and 3.  This has nothing to
 *  do with the process rank (above), which is a group-wide info.
 *  Initialized by gamma_init_stub(), reset by gamma_exit().
 */
static int my_procnum = 0;

/*
 *  Total no. of process instances in the group.
 *  Initialized by gamma_init_stub(), reset by gamma_exit().
 */
static int how_many_inst = 0;

/*
 *  Total no. of workstations attached to the cluster.
 *  Initialized by gamma_getconfig().
 */
static int gamma_physical_nodes = 0;

/*
 *  Physical rank in the cluster: a number from 0 to gamma_physical_nodes-1.
 *  Initialized by gamma_getconfig().
 */
static int gamma_phys_whoami = -1;

/*
 *  During execution of a receiver handler,
 *  this is the port that launched the handler itself.
 */
int gamma_active_port = 0;

/*
 *  The length of the last arrived message.
 *  Set by the GAMMA driver from kernel level upon msg arrival.
 */
int gamma_msglen = 0;

/*
 *  Clock rate of the underlying CPU.
 *  Initialized by gamma_init_stub().
 */
unsigned long gamma_clock;

/*
 *  host names and MAC addresses of the workstations in the cluster,
 *  initialized by gamma_getconfig()
 */
static char host_table[MAX_NUM_NODES][64];
static eth_addr eth[MAX_NUM_NODES];

#ifdef MORE_NICS_SAMENODE_ROUTING
/*
 *  for each entry in eth[], the nic to use to reach that host.
 */
static unsigned short nic[MAX_NUM_NODES];
#endif

#ifdef USE_FAST_BARRIER_HALFDUPLEX
/*
 *  Length of synchronization chains.
 *  Used by gamma_sync() on half duplex.
 *  Initialized by gamma_init().
 */
int proc_chain_len;
#endif

/*
 *  Keep the mapping instance # --> node name.
 *  Each such mapping is allocated and initialized by gamma_init_nospawn()
 *  and deallocated by gamma_exit_nospawn().
 */
typedef char node_name[64];
node_name *i2name_table;


/*
 *  GAMMA library functions, some of which contain
 *  "int $0x21" to trap into the GAMMA kernel code
 *  (see macros TRAP_P* in include file).
 *  FORTRAN stubs have an "f" suffix and follow the
 *  "one-or-two-underscore" naming convention.
 */


/***************************************************************************/
/*		Support routines.					   */
/***************************************************************************/


static inline void gamma_spinlock_init
(void)
{
	gamma_spinlock.s = 0;
	return;
}


static inline void gamma_spinlock_take
(void)
{

	int oldval;


	for (;;) {
		while (gamma_spinlock.s);
		asm (
			"lock; xchgl %0,%1"
			: "=q" (oldval), "=m" (gamma_spinlock.s)
			: "0" (1)
			: "memory"
		);
		if (!oldval) break;
	}
	return;
}


static inline void gamma_spinlock_release
(void)
{
	gamma_spinlock.s = 0;
	return;
}


/***************************************************************************/
/*		Send routines						   */
/***************************************************************************/


/************************************************************************
 *	Blocking send routines, best-effort (not collision-safe)	*
 ************************************************************************/

int gamma_send
(unsigned short out_port, void *data, unsigned long len)
{

	long res;


	for (;;) {
TRAP_P5(GAMMA_SEND,res,out_port,my_par_pid,my_procnum,data,len);
		if (res != -GAMMA_ERR_TRYAGAIN) break;
	}
	return (int)res;

}


int fgamma_send__
(unsigned short *out_port, void *data, unsigned long *len, int *res)
{
	*res = gamma_send(*out_port,data,*len);
	return 0;
}


int gamma_send_flowctl
(unsigned short out_port, void *data, unsigned long len)

{
	long out_port_and_was_susp, res;
	unsigned long was_suspended = 0;


	for (;;) {
		out_port_and_was_susp =
			((unsigned long)was_suspended) |
			((unsigned long)out_port);
TRAP_P5(GAMMA_SEND_FLOWCTL,res,out_port_and_was_susp,my_par_pid,my_procnum,data,len);
		if (res != -GAMMA_ERR_TRYAGAIN)
			if (res <= 0)
				return (int)res;  /*  Finished or failed  */

		/*  Brief waiting, poll for incoming (flow-ctl) messages.  */
		gamma_poll();

		if (res < 0) {
/*  Could not send anything new because of no credit.  */
#ifdef SPIN_YIELD
/*  Give chance to other processes on same node.  */
			sched_yield();
#endif
			continue;
		}

		data = (void *)(((char *)data)+(len-res));
		len = res;
		was_suspended = 0x80000000;
	}

}


int fgamma_send_flowctl__
(unsigned short *out_port, void *data, unsigned long *len, int *res)
{
	*res = gamma_send_flowctl(*out_port,data,*len);
	return 0;
}


int gamma_send_2p
(
	unsigned short out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2
)
{
/*
 *  THIS FUNCTION ROUNDS "len1" (THE SIZE OF 1ST PIECE OF DATA) TO
 *  DOUBLEWORDS.
 */

	long out_port_and_len1, my_par_pid_and_procnum, res;


	out_port_and_len1 =
		(((unsigned long)out_port)<<8) |
		(((unsigned long)(len1)+3)&(~(unsigned long)3));
	my_par_pid_and_procnum =
		((unsigned long)my_par_pid) |
		(((unsigned long)my_procnum) << 16);
	for (;;) {
TRAP_P5(GAMMA_SEND_2P,res,out_port_and_len1,my_par_pid_and_procnum,data1,data2,len2);
		if (res != -GAMMA_ERR_TRYAGAIN) break;
	}
        return (int)res;

}


int gamma_send_2p_flowctl
(
	unsigned short out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2
)
{
/*
 *  THIS FUNCTION ROUNDS "len1" (THE SIZE OF 1ST PIECE OF DATA) TO
 *  DOUBLEWORDS.
 */

	unsigned long llen2;
	long out_port_and_len1, my_par_pid_and_procnum, res;
	unsigned long was_suspended = 0;


	llen2 = len2;
	for (;;) {
		out_port_and_len1 =
			((unsigned long)was_suspended) |
			(((unsigned long)out_port)<<8) |
			(((unsigned long)(len1)+3)&(~(unsigned long)3));
		my_par_pid_and_procnum =
			((unsigned long)my_par_pid) |
			(((unsigned long)my_procnum) << 16);
TRAP_P5(GAMMA_SEND_2P_FLOWCTL,res,out_port_and_len1,my_par_pid_and_procnum,data2,len2,data1);
		if (res != -GAMMA_ERR_TRYAGAIN)
			if (res <= 0)
				return (int)res;  /*  Finished or failed  */

		/*  Brief waiting, poll for incoming (flow-ctl) messages.  */
		gamma_poll();

		if (res < 0) {
/*  Could not send anything new because of no credit.  */
#ifdef SPIN_YIELD
/*  Give chance to other processes on same node.  */
			sched_yield();
#endif
			continue;
		}

		if ((unsigned long)res > llen2) {
			/*  Suspended while sending 1st piece  */
			data1 =
			(void *)(((char *)data1)+(len1+len2-res));
			len1 = res-len2;
		}
		else {
			/*  Suspended while sending 2nd piece  */
			data1 = NULL;
			len1 = 0;
			data2 = (void *)(((char *)data2)+(len2-res));
			len2 = res;
		}
		was_suspended = 0x80000000;
	}

}


/************************************************************************
 *	Non-blocking send routines, best-effort (not collision-safe)	*
 ************************************************************************/


int gamma_isend
(unsigned short out_port, void *data, unsigned long len)
{

	long res;


	for (;;) {
TRAP_P5(GAMMA_ISEND,res,out_port,my_par_pid,my_procnum,data,len);
		if (res != -GAMMA_ERR_TRYAGAIN) break;
	}
	return (int)res;

}


int fgamma_isend__
(unsigned short *out_port, void *data, unsigned long *len, int *res)
{
	*res = gamma_isend(*out_port,data,*len);
	return 0;
}


int gamma_isend_flowctl
(unsigned short out_port, void *data, unsigned long len)
{

	long out_port_and_was_susp, res;
	unsigned long was_suspended = 0;


	for (;;) {
		out_port_and_was_susp =
			((unsigned long)was_suspended) |
			((unsigned long)out_port);
TRAP_P5(GAMMA_ISEND_FLOWCTL,res,out_port_and_was_susp,my_par_pid,my_procnum,data,len);
		if (res >= 0)
			return (int)res;  /* Started */
		if (res == -GAMMA_ISEND_IS_DONE)
			return (int)res;  /* Started and finished */
		if (res > -GAMMA_ERR_MAX)
			return (int)res;  /* Failed */
		/*
		 *  The operation has been suspended, "-(res+GAMMA_ERR_MAX)"
		 *  carries the lenght of the residual, not yet sent part of
		 *  msg.
		 */

		/*  Brief waiting, poll for incoming (flow-ctl) messages.  */
		gamma_poll();

		res = -(res+GAMMA_ERR_MAX);
		if ((unsigned long)res == len) {
/*  Could not send anything new because of no credit.  */
#ifdef SPIN_YIELD
/*  Give chance to other processes on same node.  */
			sched_yield();
#endif
			continue;
		}
		data = (void *)(((char *)data)+(len-res));
		was_suspended = 0x80000000;
		len = (unsigned long)res;
	}

}


int fgamma_isend_flowctl__
(unsigned short *out_port, void *data, unsigned long *len, int *res)
{
	*res = gamma_send_flowctl(*out_port,data,*len);
	return 0;
}


int gamma_isend_2p
(
	unsigned short out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2
)
{
/*
 *  THIS FUNCTION ROUNDS "len1" (THE SIZE OF 1ST PIECE OF DATA) TO
 *  DOUBLEWORDS.
 */

	long out_port_and_len1, my_par_pid_and_procnum, res;


	out_port_and_len1 =
		(((unsigned long)out_port)<<8) |
		(((unsigned long)(len1)+3)&(~(unsigned long)3));
	my_par_pid_and_procnum =
		((unsigned long)my_par_pid) |
		(((unsigned long)my_procnum) << 16);
	for (;;) {
TRAP_P5(GAMMA_ISEND_2P,res,out_port_and_len1,my_par_pid_and_procnum,data1,data2,len2);
		if (res != -GAMMA_ERR_TRYAGAIN) break;
	}
        return (int)res;

}


int gamma_isend_2p_flowctl
(
	unsigned short out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2
)
{
/*
 *  THIS FUNCTION ROUNDS "len1" (THE SIZE OF 1ST PIECE OF DATA) TO
 *  DOUBLEWORDS.
 */

	unsigned long llen2, llen;
	long out_port_and_len1, my_par_pid_and_procnum, res;
	unsigned long was_suspended = 0;


	llen2 = len2;
	llen = len1 + len2;
	for (;;) {
		out_port_and_len1 =
			((unsigned long)was_suspended) |
			(((unsigned long)out_port)<<8) |
			(((unsigned long)(len1)+3)&(~(unsigned long)3));
		my_par_pid_and_procnum =
			((unsigned long)my_par_pid) |
			(((unsigned long)my_procnum) << 16);
TRAP_P5(GAMMA_ISEND_2P_FLOWCTL,res,out_port_and_len1,my_par_pid_and_procnum,data2,len2,data1);
		if (res >= 0)
			return (int)res;  /* Started */
		if (res == -GAMMA_ISEND_IS_DONE)
			return (int)res;  /* Started and finished */
		if (res > -GAMMA_ERR_MAX)
			return (int)res;  /* Failed */
		/*
		 *  The operation has been suspended, "-(res+GAMMA_ERR_MAX)"
		 *  carries the lenght of the residual, not yet sent part of
		 *  msg.
		 */

		/*  Brief waiting, poll for incoming (flow-ctl) messages.  */
		gamma_poll();

		res = -(res+GAMMA_ERR_MAX);
		if ((unsigned long)res == llen) {
/*  Could not send anything new because of no credit.  */
#ifdef SPIN_YIELD
/*  Give chance to other processes on same node.  */
			sched_yield();
#endif
			continue;
		}
		was_suspended = 0x80000000;
		if ((unsigned long)res > llen2) {
			/*  Suspended while sending 1st piece  */
			data1 =
			(void *)(((char *)data1)+(len1+len2-res));
			len1 = res-len2;
		}
		else {
			/*  Suspended while sending 2nd piece  */
			data1 = NULL;
			len1 = 0;
			data2 = (void *)(((char *)data2)+(len2-res));
			len2 = res;
		}
	}

}


int gamma_wsend
(int isend_handle)
{

	long res;


	if (isend_handle == -GAMMA_ISEND_IS_DONE) return 0;
	TRAP_P1(GAMMA_WSEND,res,isend_handle);
	return (int)res;

}


int fgamma_wsend__
(int *isend_handle, int *res)
{
	*res = gamma_wsend(*isend_handle);
	return 0;
}


int gamma_tsend
(int isend_handle)
{

	long res;


	if (isend_handle == -GAMMA_ISEND_IS_DONE) return 0;
	TRAP_P1(GAMMA_TSEND,res,isend_handle);
	return (int)res;

}


/***************************************************************************/
/*		Receive routines.					   */
/***************************************************************************/


static inline int post_polling
(unsigned long poll_res)
{

	unsigned long port;


	port = poll_res & PORT_MASK;
	if (!(poll_res & IS_FIRST_CHUNK)) {
		/*  Whole msg complete  */
		if (likely(gamma_port[port].in_use)) {
			sync_sem1[gamma_port[port].sem_idx]++;
			if (likely(gamma_port[port].rec_handler == NULL))
				return (int)port;
			if (likely(gamma_port[port].handler_has_run)) {
				gamma_port[port].handler_has_run = 0;
				return (int)port;
			}
			gamma_active_port = port;
			/*  Invoke handler  */
			gamma_port[port].rec_handler();
			return (int)port;
		}
		return -GAMMA_ERR_GENERIC;
	}
	/*  1st message chunk complete  */
	if (likely(gamma_port[port].in_use)) {
		if (unlikely(gamma_port[port].rec_handler == NULL))
			return -GAMMA_RX_INCOMPLETE;
		if (unlikely(gamma_port[port].handler_has_run))
			return -GAMMA_RX_INCOMPLETE;
		gamma_active_port = port;
		/*  Invoke handler  */
		gamma_port[port].rec_handler();
		gamma_port[port].handler_has_run = 1;
		return (int)port;
	}
	return -GAMMA_ERR_GENERIC;
}


/*  Returns the port of a new msg arrival, if any; otherwise < 0  */
inline int gamma_poll
(void)
{

	long res,res1;


	gamma_spinlock_take();  /*  run rx() and rec handler atomically  */
	for (;;) {
		TRAP_P2(GAMMA_POLL,res,my_par_pid,my_procnum);

#ifdef MORE_INSTANCES_ON_SAME_NODE
sigh:
#endif
		if (res < 0) break;
		res1 = post_polling(res);
		if (res & IS_FIRST_CHUNK) {
			/* it was 1st chunk of a msg, go take remaining part */
#ifdef MORE_INSTANCES_ON_SAME_NODE
			if (res & IS_INTRA_NODE) {
				/* an intra node msg, so poll local */
				TRAP_P2(GAMMA_POLL_INTRANODE,res,
					my_par_pid,my_procnum);
				goto sigh;
			}
#endif
			continue;
		}
		gamma_spinlock_release();
		return res1;
	}
	gamma_spinlock_release();
	if (res == -GAMMA_ERR_MUSTYIELD)
		sched_yield();
	return (int)res;

}


/*  Returns the port of a new msg arrival, if any; otherwise < 0  */
inline int gamma_poll_nowait
(void)
{

	long res,res1;


	gamma_spinlock_take();  /*  run rx() and rec handler atomically  */
	for (;;) {
		TRAP_P2(GAMMA_POLL_NOWAIT,res,my_par_pid,my_procnum);

#ifdef MORE_INSTANCES_ON_SAME_NODE
sigh:
#endif
		if (res < 0) break;
		res1 = post_polling(res);
		if (res & IS_FIRST_CHUNK) {
			/* it was 1st chunk of a msg, go take remaining part */
#ifdef MORE_INSTANCES_ON_SAME_NODE
			if (res & IS_INTRA_NODE) {
				/* an intra node msg, so poll local */
				TRAP_P2(GAMMA_POLL_INTRANODE,res,
					my_par_pid,my_procnum);
				goto sigh;
			}
#endif
			continue;
		}
		gamma_spinlock_release();
		return res1;
	}
	gamma_spinlock_release();
	if (res == -GAMMA_ERR_MUSTYIELD)
		sched_yield();
	return (int)res;

}


int gamma_signal
(unsigned short sem)
{
	/*
	 *  Increment semaphore "sem"...
	 */
	sync_sem1[sem]++;
#ifdef USE_DAMNICK
	pre_sync_sem[sem]++;
#endif
	return 0;
}


int fgamma_signal__
(unsigned short *sem)
{
	return gamma_signal(*sem);
}


int gamma_sigerr
(unsigned short sem)
{
	/*
	 *  Increment the error counter of semaphore "sem" by one.
	 */
	sync_err[sem]++;
	return 0;
}


int fgamma_sigerr__
(unsigned short *sem, int *res)
{
	return gamma_sigerr(*sem);
}


int gamma_wait
(unsigned short sem, unsigned long n)
{
/*
 *  Implemented as a spinning on semaphore "sem"
 *  while polling the NIC's Rx ring.
 *  Return when either gamma_signal(sem) has been issued at least
 *  "n" times, or when gamma_sigerr(sem) has been issued at least once.
 *  Upon return with no error, decrement the current value of "sem" by "n".
 *  Upon return with error, reset the error counter of "sem".
 */

unsigned long diff;
#ifdef SPIN_YIELD
time_586 a,b;
unsigned long yield_phase = 0;
double spin_time;
#endif
int err;


if (sync_sem1[sem] < sync_sem2[sem]) goto wrapped_around;
diff = sync_sem1[sem]-sync_sem2[sem];

diff1:

if (diff < n) {
#ifdef SPIN_YIELD
	gamma_time(a);
#endif
	for (;;) {
		/*  Check the current error count of "sem"  */
		if ( sync_err[sem] ) goto error;
#ifdef USE_DAMNICK
		if (pre_sync_sem[sem] != sync_sem1[sem]) {
/*
 *  New message hit this semaphore.  Wait for completion of
 *  notification (the DAMNICK device is in charge of this).
 *  Notification is complete (and so is msg arrival) when
 *  the two variables hold the same value.
 */
			time_586 aa,bb;
			gamma_time(aa);
			for (;;) {
				if (pre_sync_sem[sem] == sync_sem1[sem]) break;
				gamma_time(bb);
				if (gamma_time_diff(bb,aa) > 100) break;
			}
		}
		else
			gamma_poll();
#else
		gamma_poll();
#endif
		if (sync_sem1[sem] < sync_sem2[sem]) goto wrapped_around2;
		diff = sync_sem1[sem]-sync_sem2[sem];

diff2:

		if (diff >= n) break;

#ifdef SPIN_YIELD
		if (yield_phase == 0) {
			/*  Busywaiting loop  */
			gamma_time(b);
			spin_time = gamma_time_diff(b,a);
			if (spin_time < SPIN_TIME)
				continue;
			/*
			 *  Time of busywaiting loop exhausted.
			 *  Switch to test-yield loop.
			 */
			 yield_phase = 1;
		}
		/*  "yield" part of the test-yield loop  */
		sched_yield();
#endif
		continue;

wrapped_around2:

		/*
		 *  sync_sem1[sem] has wrapped around
		 *  while sync_sem2[sem] has not.
		 */
		diff = (unsigned long)
			((double)sync_sem1[sem]
			+ (double)(1<<16)*(double)(1<<16)
			- (double)sync_sem2[sem]);
		goto diff2;
	}
}

/*
 *  Decrement semaphore "sem"...
 */
sync_sem2[sem] += n;
return 0;

error:

err = sync_err[sem];
sync_err[sem] = 0;
return -err;

wrapped_around:

/*  sync_sem1[sem] has wrapped around while sync_sem2[sem] has not  */
diff = (unsigned long)
	((double)sync_sem1[sem]
	+ (double)(1<<16)*(double)(1<<16)
	- (double)sync_sem2[sem]);
goto diff1;

}


int fgamma_wait__
(unsigned short *sem, unsigned long *n, int *res)
{
	*res = gamma_wait(*sem,*n);
	return 0;
}


int gamma_test
(unsigned short sem)
{
/*
 *  Return the current value (always >= 0) of semaphore "sem", if no errors
 *  occurred.
 *  Otherwise, the error code (< 0).
 */

	int res;


	gamma_poll_nowait();

	if (sync_err[sem]) goto err;

	if (sync_sem1[sem] >= sync_sem2[sem])
		res = (unsigned long)(sync_sem1[sem] - sync_sem2[sem]);
	else
		res =	((double)sync_sem1[sem]
			+ (double)(1<<16)*(double)(1<<16)
			- (double)sync_sem2[sem]);

	return res;

err:

	res = sync_err[sem];
	sync_err[sem] = 0;
	return -res;

}


int fgamma_test__
(unsigned short *sem, int *res)
{
	*res = gamma_test(*sem);
	return 0;
}


int gamma_suspend_irq
(void)
{
	long res;
	TRAP_P0(GAMMA_SUSPEND_IRQ,res);
	return (int)res;
}


int gamma_resume_irq
(void)
{
	long res;
	TRAP_P0(GAMMA_RESUME_IRQ,res);
	return (int)res;
}


/***************************************************************************/
/*		Port programming routines.				   */
/***************************************************************************/


static inline int gamma_connect_port_stub
(
	unsigned long par_pid_and_port_and_procnum,
	unsigned short partner_instnum,
	unsigned long partner_port_and_partner_par_pid,
	unsigned short semaphore,
	unsigned short buffer_kind_and_boundary_1p
)
{
	long res;
TRAP_P5(GAMMA_CONNECTPORT,res,par_pid_and_port_and_procnum,partner_instnum,partner_port_and_partner_par_pid,semaphore,buffer_kind_and_boundary_1p);
	return (int)res;
}


static inline int gamma_connect_port
(
	unsigned short port,
	unsigned short partner_instnum,
	unsigned char partner_par_pid,
	unsigned short partner_port,
	void (*handler_rec)(),
	unsigned short semaphore,
	unsigned char buffer_kind,
	unsigned char boundary_1p
)
{
	/*
	 *  Connect a port to a destination,
	 *  bind the port to recv handler,
	 *  bind the port to a notification semaphore,
	 *  set the ``kind'' of recv buffer (see include files).
	 */

	short buffer_kind_and_boundary_1p;


        if (
		(!my_par_pid)
		| (port > MAX_PORT_NUMBER)
		| ((partner_instnum > MAX_NUM_NODES*MAX_NUM_LOCAL_INST)
			&& (partner_instnum != BROADCAST))
//		| (partner_par_pid > MAX_PAR_PID)
		| (partner_port > MAX_PORT_NUMBER)
        )
                return -GAMMA_ERR_GENERIC;

	/* We assume a handler will not span more than a page */
	gamma_port[port].sem_idx = semaphore;
	gamma_port[port].in_use = 1;
	gamma_port[port].handler_has_run = 0;
	gamma_port[port].rec_handler = handler_rec;
	sync_sem1[semaphore] = 0;
	sync_sem2[semaphore] = 0;
	pre_sync_sem[semaphore] = 0;
	sync_err[semaphore] = 0;
	buffer_kind_and_boundary_1p =
		(((unsigned short)buffer_kind) << 8)
		| ((unsigned short)boundary_1p);
	return gamma_connect_port_stub(
		(((unsigned long)my_par_pid) << 24)
		| (((unsigned long)port) << 8)
		| ((unsigned long)my_procnum),
		partner_instnum,
		(((unsigned long)partner_port) << 8)
		| ((unsigned long)partner_par_pid),
		semaphore,
		buffer_kind_and_boundary_1p
	);

}


int gamma_set_passive_port
(
	unsigned short port,
	unsigned short partner_instnum,
	unsigned char partner_par_pid,
	unsigned short partner_port,
	unsigned short semaphore,
	unsigned char buffer_kind,
	void *buffer_in,
	unsigned long buffer_in_len
)
{
	/*
	 *  Connect a port and attach a recv buffer (if any).
	 *  Will NOT use recv handler (``passive port''), but
	 *  will use a semaphore to notify message arrivals.
	 */

	long res;


	res = gamma_connect_port(
		port,partner_instnum,partner_par_pid,partner_port,NULL,
		semaphore,buffer_kind,0
	);
	if (res < 0) return res;
	if (buffer_in_len != 0)
		res = gamma_post_recv(port,buffer_in,buffer_in_len);
	return res;

}


int fgamma_set_passive_port__
(
	unsigned short *port,
	unsigned short *partner_instnum,
	unsigned char *partner_par_pid,
	unsigned short *partner_port,
	unsigned short *semaphore,
	unsigned char *buffer_kind,
	void *buffer_in,
	unsigned long *buffer_in_len,
	int *res
)
{
	*res = gamma_set_passive_port(
		*port,*partner_instnum,*partner_par_pid,*partner_port,
		*semaphore,*buffer_kind,buffer_in,*buffer_in_len);
	return 0;
}


int gamma_set_active_port
(
	unsigned short port,
	unsigned short partner_instnum,
	unsigned char partner_par_pid,
	unsigned short partner_port,
	void (*handler_rec)(),
	unsigned short semaphore,
	unsigned char buffer_kind,
	void *buffer_in,
	unsigned long buffer_in_len
)
{
	/*
	 *  Connect a port and attach a recv buffer (if any).
	 *  Will use recv handler (``active port''), in addition to
	 *  using a notification semaphore.
	 */

	long res;


	res = gamma_connect_port(
		port,partner_instnum,partner_par_pid,partner_port,handler_rec,
		semaphore,buffer_kind,0
	);
	if (res < 0) return res;
	if (buffer_in_len != 0)
		res = gamma_post_recv(port,buffer_in,buffer_in_len);
	return res;

}


int fgamma_set_active_port__
(
	unsigned short *port,
	unsigned short *partner_instnum,
	unsigned char *partner_par_pid,
	unsigned short *partner_port,
	void (*handler_rec)(),
	unsigned short *semaphore,
	unsigned char *buffer_kind,
	void *buffer_in,
	unsigned long *buffer_in_len,
	int *res
)
{
	*res = gamma_set_active_port(
		*port,*partner_instnum,*partner_par_pid,*partner_port,
		handler_rec,*semaphore,*buffer_kind,buffer_in,*buffer_in_len);
	return 0;
}


int gamma_set_active_port_2p
(
	unsigned short port,
	unsigned short partner_instnum,
	unsigned char partner_par_pid,
	unsigned short partner_port,
	void (*handler_rec)(),
	unsigned short semaphore,
	unsigned char buffer_kind,
	void *buffer_in,
	unsigned long buffer_in_len,
	unsigned char boundary_1p
)
{
	/*
	 *  Connect a port and attach a recv buffer (if any).
	 *  Will use recv handler (``active port''), in addition to
	 *  using a notification semaphore.
	 *  The semantics is different from usual ports:
	 *  The recv handler, if present, will run upon filling the
	 *  first "boundary_1p" bytes of the destination buffer,
	 *  rather than at the end of message.  However, the notification
	 *  on the user-provided GAMMA semaphore will occur at message end
	 *  as usual.
	 */

	long res;


	res = gamma_connect_port(
		port,partner_instnum,partner_par_pid,partner_port,handler_rec,
		semaphore,buffer_kind,boundary_1p
	);
	if (res < 0) return res;
	if (buffer_in_len != 0)
		res = gamma_post_recv(port,buffer_in,buffer_in_len);
	return res;

}


int gamma_post_recv
(unsigned short port, void *buffer_in, unsigned long buffer_in_len)
{
	/*
	 *  Attach a valid user-space input buffer to a port for immediate use.
	 */

	long res;
	if (port > MAX_PORT_NUMBER) return -GAMMA_ERR_GENERIC;
TRAP_P5(GAMMA_POST_RECV,res,my_par_pid,my_procnum,port,buffer_in,buffer_in_len);
	return res;
}


int fgamma_post_recv__
(unsigned short *port, void *buffer_in, unsigned long buffer_len, int *res)
{
	*res = gamma_post_recv(*port,buffer_in,buffer_len);
	return 0;
}


int gamma_switch_to_buffer
(unsigned short port, void *buffer_in, unsigned long buffer_in_len)
{
	/*
	 *  Attach a valid user-space input buffer to a port
	 *  and use it for the remaining part of the
	 *  current message ONLY
	 */

	long res;
	if (port > MAX_PORT_NUMBER) return -GAMMA_ERR_GENERIC;
TRAP_P5(GAMMA_SWITCH_TO_BUFFER,res,my_par_pid,my_procnum,port,buffer_in,buffer_in_len);
	return res;
}


int fgamma_switch_to_buffer__
(unsigned short *port, void *buffer_in, unsigned long buffer_len, int *res)
{
	*res = gamma_switch_to_buffer(*port,buffer_in,buffer_len);
	return 0;
}


void *gamma_recvptr
(unsigned short port)
{
	/*
	 *  Returns the position, in the currently attached receive
	 *  buffer, where next message is expected to be copied.
	 */

	void * res;
	if (port > MAX_PORT_NUMBER) return (void *)-GAMMA_ERR_GENERIC;
TRAP_P3(GAMMA_RECVPTR,res,my_par_pid,my_procnum,port);
	return res;
}


/***************************************************************************/
/*		Init/Exit routines.					   */
/***************************************************************************/


static void gamma_lock_internal_data
(void)
{
	/*
	 *  Lock all library data structures that may be accessed by the
	 *  GAMMA driver from kernel side
	 */
#ifdef USE_DAMNICK
	gamma_mlock((void *)sync_sem1,
		(MAX_SYNC_SEM_NUMBER+1)*sizeof(unsigned long));
	gamma_mlock((void *)pre_sync_sem,
		(MAX_SYNC_SEM_NUMBER+1)*sizeof(unsigned long));
#endif
	gamma_mlock(&gamma_msglen,sizeof(int));
}


static void gamma_set_internal_ports_frontend
(void)
{
	/*
	 *  Setting communication ports and handlers
	 *  for collective communication routines
	 *  (front-end instance).
	 */
	gamma_set_passive_port(SYNC_PORT,
		BROADCAST,my_par_pid,SYNC_PORT,
		SYNC_PORT,GO_AHEAD,NULL,0);
	gamma_set_passive_port(BCAST_PORT,
		BROADCAST,my_par_pid,BCAST_PORT,
		BCAST_PORT,GO_AHEAD,NULL,0);
}


static void gamma_set_internal_ports_slaves
(void)
{
	/*
	 *  Setting communication ports and handlers
	 *  for collective communication routines.
	 */
#ifdef USE_FAST_BARRIER_HALFDUPLEX
	/*
	 *  Set ports for the two-chains barrier
	 *  optimized for repeater hubs.
	 */
	if (proc_chain_len == 1)
		gamma_set_passive_port(
			SYNC_PORT,0,my_par_pid,SYNC_PORT,
			SYNC_PORT,GO_AHEAD,NULL,0
		);
	else
		if ((my_instnum % proc_chain_len) == 1)
			gamma_set_passive_port(SYNC_PORT, 
				0,my_par_pid,SYNC_PORT,
				SYNC_PORT,GO_AHEAD,NULL,0
			);
		else
			gamma_set_passive_port(SYNC_PORT, 
				my_instnum-1,my_par_pid,SYNC_PORT,
				SYNC_PORT,GO_AHEAD,NULL,0
			);
#else
	/*  Set ports for a traditional tree barrier  */

	int depth,i;


	for (i=MAX_NUM_NODES*MAX_NUM_LOCAL_INST, depth=-1; i; i>>=1, depth++);

	for (i=1;i<=depth;i++)
		if (((my_instnum >> i) << i) != my_instnum) break;
	gamma_set_passive_port(SYNC_PORT, 
		my_instnum - (1<<(i-1)), my_par_pid, SYNC_PORT,
		SYNC_PORT,GO_AHEAD,NULL,0
	);
#endif
	gamma_set_passive_port(BCAST_PORT,
		BROADCAST,my_par_pid,BCAST_PORT,
		BCAST_PORT,GO_AHEAD,NULL,0
	);

}


static void gammasighandler_poll
(int sigval)
{
	if (gamma_spinlock.s) return;
	gamma_poll();
}


static void gammasighandler_abort
(int sigval)
{
	fprintf(stderr,
	"GAMMA: signal %d: process instance %d aborting on node %s...\n",
		sigval,my_instnum,host_table[gamma_phys_whoami]);
	gamma_abort();
	abort();
}


static int find_in_argv
(int argc, char **argv, char *string)
{
	int i;
	if (argc < 2) return 0;
	for (i=1; i<argc-1; i++)
		if (!strcmp(argv[i],string)) return i;
	return 0;
}


static int
gamma_vm_create
(
	u_int32_t rnd_key,
	unsigned int num_instances,
	void *node_list_in, void *node_list_out
)
{
	long res;
	TRAP_P4(GAMMA_VM_CREATE,res,rnd_key,num_instances,node_list_in,
		node_list_out);
	my_rnd_key = rnd_key;
	return (int)res;  /*  the par_pid of the created virtual GAMMA  */
}


static int gamma_init_stub
(
	void *gamma_msglen_ptr,
	void *sem1_table_ptr,
	void *pre_sem_table_ptr
)
{

	long res,res1,res2;


TRAP_P1(GAMMA_INIT_PART1,res,gamma_msglen_ptr);
TRAP_P2(GAMMA_INIT_PART2,res1,sem1_table_ptr,pre_sem_table_ptr);
TRAP_P0(GAMMA_INIT_PART3,res2);
	my_instnum = (unsigned short)res;
	how_many_inst = (unsigned short)(res >> 16);
	my_procnum = (unsigned short)res1;
	my_par_pid = (unsigned short)(res1 >> 16);
	gamma_clock = (unsigned long)res2;
	signal(SIGURG,gammasighandler_poll);
	signal(SIGTERM,gammasighandler_abort);
	signal(SIGINT,gammasighandler_abort);
	signal(SIGPIPE,gammasighandler_abort);
	signal(SIGSEGV,gammasighandler_abort);
	signal(SIGFPE,gammasighandler_abort);
	signal(SIGQUIT,gammasighandler_abort);
	return (int)res;

}


int gamma_init
(unsigned int num_instances, int argc, char **argv)
{

int 		i, j, jj, pid, res, par_pid, mach_num_nodes;
unsigned short	node_list[MAX_NUM_NODES*MAX_NUM_LOCAL_INST];
char		*rsharg[30],
		pwd[MAX_NAME_LEN],
		progname[MAX_NAME_LEN],
		gammakey_string[16],
		key_string[16],
		gammahome_string[16];
time_586	a;

res = gamma_getconfig(0);
if (res < 0)
	switch (res) {
		case -GAMMA_ERR_NODEV:
fprintf(stderr,"No GAMMA devices found.  Abort...\n");
			exit(1);
		case -GAMMA_ERR_NOIFCONFIG:
fprintf(stderr,"Must ifconfig the GAMMA NICs before use.  Abort...\n");
			exit(1);
		case -GAMMA_ERR_NOGAMMACONF:
fprintf(stderr,"Inconsistent /etc/gamma.conf file.  Abort...\n");
			exit(1);
		case -GAMMA_ERR_NOINIT:
fprintf(stderr,"Must run \"gammagetconfig\" before using GAMMA.  Abort...\n");
			exit(1);
		default:
fprintf(stderr,"gamma_init():  Generic error.  Abort...\n");
			exit(1);
	}

/*  is num_instances out of range?  */	
if ((num_instances < 1)
	|| (num_instances > gamma_physical_nodes * MAX_NUM_LOCAL_INST)) {
	fprintf(stderr,
	"gamma_init(): wrong or excessive number of instances.  Abort...\n");
	exit(1);
}

/*  Init the port table  */
for (i=0;i<=MAX_PORT_NUMBER;i++) {
	gamma_port[i].in_use = 0;
	gamma_port[i].sem_idx = 0;
	gamma_port[i].rec_handler = NULL;
	gamma_port[i].handler_has_run = 0;
}

gamma_spinlock_init();

for (i=0;i<=MAX_SYNC_SEM_NUMBER;i++) {
	sync_sem1[i] = 0;
	sync_sem2[i] = 0;
	pre_sync_sem[i] = 0;
	sync_err[i] = 0;
}

/*  Check whether this process instance is master or slave.  */
/*  Look for arguments -GAMMAKEY <rnd_key> in argv  */

if ((res = find_in_argv(argc,argv,"-GAMMAKEY"))) {
	/*
	 *  This is a slave instance.
	 *  argv[res] is "-GAMMAKEY", and argv[res+1] is the rnd_key of
	 *  an already created GAMMA virtual machine.
	 */
	my_rnd_key = (u_int32_t)atoi(argv[res+1]);

	/*
	 *  We should also find additional args: "-GAMMAHOME" <pwd>
	 */
	if (!(i = find_in_argv(argc,argv,"-GAMMAHOME"))) {
fprintf(stderr,"gamma_init():  Working dir not specified on slave instance\n");
		exit(1);
	}
	if (chdir(argv[i+1])) {
fprintf(stderr,"gamma_init():  Working dir not found on slave instance\n");
		exit(1);
	}

	gamma_suspend_irq();

	/*  Register the process with the virtual GAMMA using the key  */
	par_pid = gamma_register(my_rnd_key);
	if (res < 0) {
		fprintf(stderr,"gamma_init():  registration failed\n");
		gamma_abort_(my_rnd_key);
		gamma_resume_irq();
		exit(1);
	}

	res = gamma_init_stub(
		(void *)&gamma_msglen,
		(void *)sync_sem1,
		(void *)pre_sync_sem
	);

	if (res < 0) {
		fprintf(stderr,"gamma_init() failed!\n");
		gamma_abort_(my_rnd_key);
		gamma_resume_irq();
		exit(1);
	}

#ifdef USE_FAST_BARRIER_HALFDUPLEX
	proc_chain_len =
		how_many_inst == 1 ? 1 :
		(how_many_inst-1+(SYNC_SUBSETS-1))/SYNC_SUBSETS;
#endif

	/*
	 *  Lock all library data structures that may be accessed by handlers.
	 */
	gamma_lock_internal_data();

	/*
	 *  Setting communication ports and handlers
	 *  for collective communication routines.
	 */
	gamma_set_internal_ports_slaves();

	/*
	 *  Temporarily setting communication ports and handlers
	 *  for safe barrier sync.
	 *  Safe barrier is a plain concurrent one.
	 */
#ifndef SYNC_SAFE_BCAST_SW
	gamma_set_passive_port(0,0,my_par_pid,0,0,GO_AHEAD,NULL,0);
#else
	gamma_set_passive_port(0,0,my_par_pid,gamma_physical_nodes,0,GO_AHEAD,NULL,0);
#endif

	gamma_sync_safe();

	return 0;
}

/*  This process is master.  Create a virtual GAMMA.  */

mach_num_nodes = 0;

/*  Look for arguments "-machinefile" <filename> in argv  */

if ((res = find_in_argv(argc,argv,"-machinefile"))) {
	char machinename[1024];
	FILE *f;

        /*  Read the user-proposed machine file.  */

	f = fopen(argv[res+1],"r");
	if (f == NULL) {
		/*  Machine file unreadable.  */
		fprintf(stderr,
			"gamma_init(): machine file unreadable.  Abort...\n");
		exit(1);
	}

	mach_num_nodes = 0;

	/*  First node of the list is always the local one.  */
	node_list[mach_num_nodes] = gamma_phys_whoami;
	mach_num_nodes++;

READ_MACHINEFILE:

	while (fscanf(f,"%s",machinename) != EOF) {
		for (i=0;i<gamma_physical_nodes;i++)
			if (!strcmp(machinename,host_table[i])) {
				node_list[mach_num_nodes] = i;
				break;
			}
		if (i >= gamma_physical_nodes) {
			/*
			 *  Machine file contains a machine not listed in
			 *  /etc/gamma.conf
			 */
			mach_num_nodes = -1;
			break;
		}
		mach_num_nodes++;
		if (mach_num_nodes == num_instances) break;
	}

	fclose(f);

	if (mach_num_nodes == 1) {
		/*  Machine file empty.  */
		fprintf(stderr,
			"gamma_init(): empty machine file.  Abort...\n");
		exit(1);
	}

	if (mach_num_nodes == -1) {
		fprintf(stderr,
"gamma_init(): machine file inconsistent with /etc/gamma.conf.  Abort...\n");
		exit(1);
	}

	if ((mach_num_nodes > 1) && (mach_num_nodes < num_instances)) {
		/*
		 *  Machine file too short (but not empty nor inconsistent
		 *  nor unreadable).
		 *  Re-read it from beginning (no risk of endless loop).
		 */
		f = fopen(argv[res+1],"r");
		goto READ_MACHINEFILE;
	}
	
}

/*
 *  At this point: if mach_num_nodes == 0 then the user did not provide
 *  any machine file; gamma_vm_create() will use the default
 *  machine list (/etc/gamma.conf) in this case.
 *  Otherwise, mach_num_nodes is equal to num_instances (which is >= 1).
 */

/*
 * Check whether we're trying to launch more process instances than allowed...
 */
if (!mach_num_nodes) {
	if (num_instances > gamma_physical_nodes * MAX_NUM_LOCAL_INST) {
		if (MAX_NUM_LOCAL_INST == 1)
			fprintf(stderr,
"gamma_init(): support for multiple process instances per processor is not enabled.  Abort...\n");
		else
			fprintf(stderr,
"gamma_init(): trying to launch more than %d process instances on at least one of the nodes.  Abort...\n",MAX_NUM_LOCAL_INST);
		exit(1);
	}
}
else {
	/* Find the max no. of process instances on a given processor */
	int n, max = -1, who = -1;
	for (i=0;i<num_instances;i++) {
		n = 1;
		for (j=i+1;j<num_instances;j++)
			if (node_list[i] == node_list[j]) n++;
		if (n > max) {
			max = n;
			who = i;
		}
	}
	if (max > MAX_NUM_LOCAL_INST) {
		if (MAX_NUM_LOCAL_INST == 1) {
			fprintf(stderr,
"gamma_init(): support for multiple process instances per node is not enabled.\n");
		}
		fprintf(stderr,
"gamma_init(): trying to launch %d process instances on node %s.\n"
"This is more than the allowed maximum of %d.\n"
"Check out the machinefile.  Abort...\n",
			max,host_table[node_list[who]],MAX_NUM_LOCAL_INST);
		exit(1);
	}
}

/* Choose a key for the GAMMA virtual machine to be created... */
gamma_time(a);
srand((unsigned int)a.lo);
my_rnd_key = (u_int32_t)rand();

if (mach_num_nodes)
	par_pid = gamma_vm_create(my_rnd_key,num_instances,&node_list,&node_list);
else
	par_pid = gamma_vm_create(my_rnd_key,num_instances,NULL,&node_list);
if (par_pid < 0) {
	fprintf(stderr,"gamma_init(): could not create a virtual GAMMA.\n");
	gamma_abort_(my_rnd_key);
	exit(1);
}

#ifdef VERBOSE
fprintf(stderr,
	"gamma_init(): created virtual GAMMA #%d (key %u) with nodes:\n",
	par_pid, my_rnd_key);
for (i=0;i<num_instances;i++)
	fprintf(stderr,"%d:  %d --> %s\n",
		i,node_list[i],host_table[node_list[i]]);
#endif

gamma_suspend_irq();

res = gamma_register(my_rnd_key);
if (res != par_pid) {
	fprintf(stderr,"gamma_init():  master: registration failed\n");
	gamma_abort_(my_rnd_key);
	gamma_resume_irq();
	exit(1);
}

res = gamma_init_stub(
	(void *)&gamma_msglen,
	(void *)sync_sem1,
	(void *)pre_sync_sem
);

if (res < 0) {
	fprintf(stderr,"gamma_init() failed!\n");
	gamma_abort_(my_rnd_key);
	gamma_resume_irq();
	exit(1);
}

#ifdef USE_FAST_BARRIER_HALFDUPLEX
proc_chain_len =
	how_many_inst == 1 ? 1 :
	(how_many_inst-1+(SYNC_SUBSETS-1))/SYNC_SUBSETS;
#endif

/*
 *  Lock all library data structures that may be accessed by handlers.
 */
gamma_lock_internal_data();

/*
 *  Setting communication ports and handlers
 *  for collective communication routines.
 */
gamma_set_internal_ports_frontend();

/*
 *  Temporarily setting communication ports and handlers
 *  for safe barrier sync (will be used only in gamma_init()).
 *  Safe barrier is a plain concurrent one.
 */
#ifndef SYNC_SAFE_BCAST_SW
gamma_set_passive_port(0,BROADCAST,my_par_pid,0,0,GO_AHEAD,NULL,0);
#else
gamma_set_passive_port(gamma_physical_nodes,BROADCAST,my_par_pid,0,gamma_physical_nodes,GO_AHEAD,NULL,0);
for(i=1; i<how_many_inst; i++)
     gamma_set_passive_port(i,i,my_par_pid,0,i,GO_AHEAD,NULL,0);
#endif

/* Be "sure" virtual GAMMA on remote hosts is set up */
gamma_busywait_usec(1000);

/*  Spawn instances of the application across the cluster  */

sprintf(key_string,"%u",my_rnd_key);
sprintf(gammakey_string,"-GAMMAKEY");
sprintf(gammahome_string,"-GAMMAHOME");
sprintf(pwd,"%s/",getenv("PWD"));

for (i=1;i<num_instances;i++) {
	pid = fork();
	if (!pid) {
		/*  child process  */
		/*  remote instances, spawning them via ssh/rsh  */

#ifdef USE_SSH
		rsharg[0] = "/usr/bin/ssh";
#else
		rsharg[0] = "/usr/bin/rsh";
#endif
		rsharg[1] = "-n";
		rsharg[2] = host_table[node_list[i]];
		if (argv[0][0] != '/') {
			/*
			 *  Application name must be completed
			 *  to the absolute name
			 */
			strcpy(progname,pwd);
			strcat(progname,argv[0]);
			rsharg[3] = progname;
		}
		else
			rsharg[3] = argv[0];


		/*  append the GAMMA-related additional args: */
		/*  -GAMMAHOME <pwd> -GAMMAKEY <par_pid>  */

		j = 4;
		for (jj = argc-1; jj; jj--) {
			rsharg[j] = argv[j-3];
			j++;
		}
		rsharg[j++] = gammahome_string;
		rsharg[j++] = pwd;
		rsharg[j++] = gammakey_string;
		rsharg[j++] = key_string;
		rsharg[j] = NULL;

#ifdef VERBOSE
		fprintf(stderr,
			"gamma_init(): launching %s on remote node %s\n",
			rsharg[3], rsharg[2]);
		fprintf(stderr,"(command:  ");
		for (jj=0;jj<=j;jj++)
			fprintf(stderr,"%s ",rsharg[jj]);
		fprintf(stderr,"\n");
#endif

		execve(rsharg[0],rsharg,NULL);
		fprintf(stderr,"gamma_init(): launch on node %s failed!\n",
			host_table[node_list[i]]);
		gamma_abort_(my_rnd_key);
		gamma_resume_irq();
		exit(1);
	}
}

gamma_sync_safe();

return 0;

}


static int gamma_vm_wish
(u_int32_t random_key, unsigned int num_instances, int is_master, int first)
{

	long res;


#if 0
	fprintf(stderr,
		"gamma_vm_wish(): key %d, num_instances %d, is_master %d...",
		random_key,num_instances,is_master);
#endif

	TRAP_P4(GAMMA_VM_WISH,res,random_key,num_instances,is_master,first);

#ifdef VERBOSE
	fprintf(stderr,"res %ld\n",res);
#endif

	return (int)res;

}


static int gamma_vm_create_bykey
(
	u_int32_t random_key,
	unsigned int num_instances,
	void *node_list
)
{
	long res;
	TRAP_P3(GAMMA_VM_CREATE_BYKEY,res,random_key,num_instances,node_list);
	return (int)res;  /*  if > 0, res is the par_pid assoc. to key  */
}


static int gamma_vm_getinfo
(void *node_list)
{

	long res;


	if (!my_par_pid) return -GAMMA_ERR_GENERIC;
	TRAP_P2(GAMMA_VM_GETINFO,res,my_par_pid,node_list);
	return (int)res;

}


int gamma_init_nospawn
(unsigned int random_key, unsigned int num_instances, int argc, char **argv)
{

int		i,res,first_time,par_pid,is_master;
unsigned short	node_list[MAX_NUM_NODES*MAX_NUM_LOCAL_INST];

res = gamma_getconfig(0);
if (res < 0)
	switch (res) {
		case -GAMMA_ERR_NODEV:
fprintf(stderr,"No GAMMA devices found.  Abort...\n");
			exit(1);
		case -GAMMA_ERR_NOIFCONFIG:
fprintf(stderr,"Must ifconfig the GAMMA NICs before use.  Abort...\n");
			exit(1);
		case -GAMMA_ERR_NOGAMMACONF:
fprintf(stderr,"Inconsistent /etc/gamma.conf file.  Abort...\n");
			exit(1);
		case -GAMMA_ERR_NOINIT:
fprintf(stderr,"Must run \"gammagetconfig\" before using GAMMA.  Abort...\n");
			exit(1);
		default:
fprintf(stderr,"gamma_init_nospawn():  Generic error.  Abort...\n");
			exit(1);
	}

/*  is num_instances out of range?  */	
if ((num_instances < 1)
	|| (num_instances > gamma_physical_nodes * MAX_NUM_LOCAL_INST)) {
		fprintf(stderr,
"gamma_init_nospawn(): wrong or excessive number of instances.  Abort...\n");
		exit(1);
}

/*  Init the port table  */
for (i=0;i<=MAX_PORT_NUMBER;i++) {
	gamma_port[i].in_use = 0;
	gamma_port[i].sem_idx = 0;
	gamma_port[i].rec_handler = NULL;
	gamma_port[i].handler_has_run = 0;
}

gamma_spinlock_init();

for (i=0;i<=MAX_SYNC_SEM_NUMBER;i++) {
	sync_sem1[i] = 0;
	sync_sem2[i] = 0;
	pre_sync_sem[i] = 0;
	sync_err[i] = 0;
}

/*
 *  Before starting the collective protocol for process grouping,
 *  make sure we will be able to interrupt it at any time in a consistent
 *  way (that is, aborting the virtual GAMMA if already created).
 */
my_rnd_key = random_key;
signal(SIGURG,gammasighandler_poll);
signal(SIGTERM,gammasighandler_abort);
signal(SIGINT,gammasighandler_abort);
signal(SIGPIPE,gammasighandler_abort);
signal(SIGSEGV,gammasighandler_abort);
signal(SIGFPE,gammasighandler_abort);
signal(SIGQUIT,gammasighandler_abort);

gamma_suspend_irq();

/*  Check whether this process instance is master or slave.  */
/*  Slaves have argument -GAMMAHOME in argv, master has not.  */

is_master = !(res = find_in_argv(argc,argv,"-GAMMAHOME"));

/*  Collective process grouping is a waiting loop  */
for (first_time = 1;;first_time = 0) {
	res = gamma_vm_wish(my_rnd_key,num_instances,is_master,first_time);
	if (!is_master)
		res = gamma_register_nospawn(my_rnd_key);
	if (res == -GAMMA_ERR_TRYAGAIN) {
		if (!is_master)
			/* don't bother with too frequent VM_WISH pkts */
			sleep(1);
		continue;
	}
	if (res >= 0) break;
	fprintf(stderr,"gamma_init_nospawn():  Error from gamma_vm_wish()\n");
	gamma_abort_(my_rnd_key);
	gamma_resume_irq();
	exit(1);
}

if (!is_master) {
	/*
	 *  This is a slave instance.  Virtual GAMMA already created,
	 *  process instance has registered.
	 */

	res = gamma_init_stub(
		(void *)&gamma_msglen,
		(void *)sync_sem1,
		(void *)pre_sync_sem
	);

	if (res < 0) {
		fprintf(stderr,"gamma_init_nospawn() failed!\n");
		gamma_abort_(my_rnd_key);
		gamma_resume_irq();
		exit(1);
	}

#ifdef USE_FAST_BARRIER_HALFDUPLEX
	proc_chain_len =
		how_many_inst == 1 ? 1 :
		(how_many_inst-1+(SYNC_SUBSETS-1))/SYNC_SUBSETS;
#endif

	/*
	 *  Lock all library data structures that may be accessed by handlers.
	 */
	gamma_lock_internal_data();

	/*
	 *  Setting communication ports and handlers
	 *  for collective communication routines.
	 */
	gamma_set_internal_ports_slaves();

	/*  Alloc the mapping instance # --> node name  */
	i2name_table = (node_name *)malloc(num_instances*sizeof(node_name));
	/*  Get the mapping instance # --> physical node # from kernel  */
	gamma_vm_getinfo(&node_list);
	/*  Init the mapping instance # --> node name  */
	for (i=0;i<num_instances;i++)
		strcpy(i2name_table[i],host_table[node_list[i]]);

	/*
	 *  Temporarily setting communication ports and handlers
	 *  for safe barrier sync.
	 *  Safe barrier is a plain concurrent one.
	 */
	gamma_set_passive_port(0,0,my_par_pid,0,0,GO_AHEAD,NULL,0);

	gamma_sync_safe();

	return 0;
}

/*  This is the master instance.  Create a virtual GAMMA.  */

par_pid = gamma_vm_create_bykey(my_rnd_key,num_instances,&node_list);
if (par_pid < 0) {
fprintf(stderr,"gamma_init_nospawn():  error from gamma_vm_create_bykey()\n");
	gamma_abort_(my_rnd_key);
	gamma_resume_irq();
	exit(1);
}

#ifdef VERBOSE
fprintf(stderr,
"gamma_init_nospawn(): created virtual GAMMA #%d (key %u) with nodes:\n",
par_pid,my_rnd_key);
for (i=0;i<num_instances;i++)
	fprintf(stderr,"%d:  %d --> %s\n",
		i,node_list[i],host_table[node_list[i]]);
#endif

res = gamma_register(my_rnd_key);
if (res != par_pid) {
	fprintf(stderr,"gamma_init_nospawn():  master: registration failed\n");
	gamma_abort_(my_rnd_key);
	gamma_resume_irq();
	exit(1);
}

res = gamma_init_stub(
	(void *)&gamma_msglen,
	(void *)sync_sem1,
	(void *)pre_sync_sem
);

if (res < 0) {
	fprintf(stderr,"gamma_init_nospawn() failed!\n");
	gamma_abort_(my_rnd_key);
	gamma_resume_irq();
	exit(1);
}

#ifdef USE_FAST_BARRIER_HALFDUPLEX
proc_chain_len =
	how_many_inst == 1 ? 1 :
	(how_many_inst-1+(SYNC_SUBSETS-1))/SYNC_SUBSETS;
#endif

/*
 *  Lock all library data structures that may be accessed by handlers.
 */
gamma_lock_internal_data();

/*
 *  Setting communication ports and handlers
 *  for collective communication routines.
 */
gamma_set_internal_ports_frontend();

/*  Alloc the mapping instance # --> node name  */
i2name_table = (node_name *)malloc(num_instances*sizeof(node_name));
/*  Get the mapping instance # --> physical node # from kernel  */
gamma_vm_getinfo(&node_list);
/*  Init the mapping instance # --> node name  */
for (i=0;i<num_instances;i++)
	strcpy(i2name_table[i],host_table[node_list[i]]);

/*
 *  Temporarily setting communication ports and handlers
 *  for safe barrier sync.
 *  Safe barrier is a plain concurrent one.
 */
gamma_set_passive_port(0,BROADCAST,my_par_pid,0,0,GO_AHEAD,NULL,0);

gamma_sync_safe();

return 0;

}


int gamma_init_argv(int *argc, char ***argv)
{

int		res,is_slave;
unsigned int	np;
u_int32_t	rnd_key;

#ifdef VERBOSE
fprintf(stderr,"gamma_init_argv(): argc %d: ",*argc);
{ int i = 0; for (; i < *argc; i++) fprintf(stderr,"|%s| ",(*argv)[i]); }
fprintf(stderr,"\n");
#endif

/*  Look for mandatory args "-GAMMANP" <no_of_process_instances> in argv  */
res = find_in_argv(*argc,*argv,"-GAMMANP");
if (!res) {
	fprintf(stderr,"gamma_init_argv(): arg -GAMMANP not found\n");
	exit(1);
}
np = (unsigned int)atoi((*argv)[res+1]);

is_slave = find_in_argv(*argc,*argv,"-GAMMAHOME");

/*  Is this call expected to spawn slave processes or not?  */

if (find_in_argv(*argc,*argv,"-GAMMANOSPAWN")) {

/*
 *  This call must not spawn remote instances; these are rather supposed to
 *  be generated by some external mechanism (e.g. Torque).
 *  Expected args (must be the last ones in the argv array):
 *  master instance: -GAMMANOSPAWN -GAMMANP <no of instances> -GAMMAKEY <key>
 *  slave instances: -GAMMANOSPAWN -GAMMANP <no of instances> -GAMMAKEY <key>
 *  			-GAMMAHOME <"pwd" where master was launched>
 */

/*  Look for argument "-GAMMAKEY" <key of virtual GAMMA>  */
	res = find_in_argv(*argc,*argv,"-GAMMAKEY");
	if (!res) {
		fprintf(stderr,"gamma_init_argv(): arg -GAMMAKEY not found\n");
		exit(1);
	}
	rnd_key = (u_int32_t)atoi((*argv)[res+1]);

#ifdef VERBOSE
	fprintf(stderr,
	"gamma_init_argv(): called with -GAMMANOSPAWN, %d instances, key %u\n",
	np,rnd_key);
#endif

	res = gamma_init_nospawn(rnd_key,np,*argc,*argv);

#ifdef STRIPARGS
	/*  Strip GAMMA-related args away  */
	if (is_slave) {
		(*argv)[*argc-7] = NULL;
		*argc -= 7;
	} else {
		(*argv)[*argc-5] = NULL;
		*argc -= 5;
	}
#endif

	return res;
}

/*  This call is expected to spawn remote instances, using rsh/ssh  */

if (!is_slave) {
	/*
	 *  Master instance.
	 *  Expected args (must be the last ones in the argv array):
	 *  -GAMMANP <no of instances> [ -machinefile <filename> ]
	 */
#ifdef VERBOSE
	fprintf(stderr,"gamma_init_argv(): master, %d instances\n",np);
#endif

	res = gamma_init(np,*argc,*argv);

#ifdef STRIPARGS
	if (find_in_argv(*argc,*argv,"-machinefile")) {
		(*argv)[*argc-4] = NULL;
		*argc -= 4;
	} else {
		(*argv)[*argc-2] = NULL;
		*argc -= 2;
	}
#endif

	return res;
}

/*
 *  Slave instance, generated by master via rsh/ssh.
 *  Expected args (must be the last ones in the argv array):
 *  -GAMMANP <no of instances> -GAMMAKEY <key>
 *  	-GAMMAHOME <"pwd" where master was launched>
 *  	[ -machinefile <filename> ]
 */
#ifdef VERBOSE
fprintf(stderr,"gamma_init_argv(): slave, %d instances\n", np);
#endif

res = gamma_init(np,*argc,*argv);

#ifdef STRIPARGS
if (find_in_argv(*argc,*argv,"-machinefile")) {
	(*argv)[*argc-8] = NULL;
	*argc -= 8;
} else {
	(*argv)[*argc-6] = NULL;
	*argc -= 6;
}
#endif

return res;

}


int fgamma_init__(int *res)
{
	int i,result;

	/*
	 *  At this point, the calling process
	 *  is already registered into a virtual GAMMA
	 */

	/*  Init the port table  */
	for (i=0;i<=MAX_PORT_NUMBER;i++) {
		gamma_port[i].in_use = 0;
		gamma_port[i].sem_idx = 0;
		gamma_port[i].rec_handler = NULL;
		gamma_port[i].handler_has_run = 0;
	}

	gamma_spinlock_init();

	for (i=0;i<=MAX_SYNC_SEM_NUMBER;i++) {
		sync_sem1[i] = 0;
		sync_sem2[i] = 0;
		pre_sync_sem[i] = 0;
		sync_err[i] = 0;
	}

	my_rnd_key = gamma_getkey();
	result = gamma_init_stub(
		(void *)&gamma_msglen,
		(void *)sync_sem1,
		(void *)pre_sync_sem
	);

	if (result < 0) {
		fprintf(stderr,"fgamma_init__() failed!\n");
		exit(1);
	}

#ifdef USE_FAST_BARRIER_HALFDUPLEX
	proc_chain_len =
		how_many_inst == 1 ? 1 :
		(how_many_inst-1+(SYNC_SUBSETS-1))/SYNC_SUBSETS;
#endif

	/*
	 *  Lock all library data structures that may be accessed by handlers.
	 */
	gamma_lock_internal_data();

	/*
	 *  Setting communication ports and handlers
	 *  for collective communication routines.
	 */
	if (!my_instnum)
		gamma_set_internal_ports_frontend();
	else
		gamma_set_internal_ports_slaves();

	/*
	 *  Temporarily setting communication ports and handlers
	 *  for safe barrier sync.
	 *  Safe barrier is a plain concurrent one.
	 */
	if (!my_instnum)
		gamma_set_passive_port(
			0,BROADCAST,my_par_pid,0,0,GO_AHEAD,NULL,0
		);
	else
		gamma_set_passive_port(
			0,0,my_par_pid,0,0,GO_AHEAD,NULL,0
		);

	gamma_suspend_irq();

	/*  To be sure enough that master exists...  */
	if (my_instnum) sleep(5);

	gamma_sync_safe();

	*res = 0;
	return 0;
}


int gamma_exit
(void)
{

	long res;
	int i,status;


	do {
		TRAP_P2(GAMMA_EXIT,res,my_par_pid,my_procnum);
	} while (res == -GAMMA_ERR_TRYAGAIN);

	if (!my_instnum)
		/*
		 *  Front-end instance has spawned num_instances-1 "slave"
		 *  instances.
		 *  Now it must wait for their death.
		 */
		for (i=1;i<how_many_inst;i++) wait(&status);

	my_par_pid = 0;
	my_instnum = 0;
	my_procnum = 0;
	how_many_inst = 0;
	res = (long)gamma_munlockall();
	gamma_resume_irq();

        return 0;

}


int gamma_exit_nospawn
(void)
{

	long res;


	do {
		TRAP_P2(GAMMA_EXIT,res,my_par_pid,my_procnum);
	} while (res == -GAMMA_ERR_TRYAGAIN);

	my_rnd_key = 0;
	my_par_pid = 0;
	my_instnum = 0;
	my_procnum = 0;
	how_many_inst = 0;
	res = (long)gamma_munlockall();
	gamma_resume_irq();

        return 0;

}


int fgamma_exit__ (int *res)
{
	*res = gamma_exit();
	return 0;
}


static int gamma_abort_
(u_int32_t rnd_key)
{
	long res;

	do TRAP_P2(GAMMA_VM_ABORT,res,rnd_key,1);
	while (res == -GAMMA_ERR_TRYAGAIN);
	return 0;

}


int gamma_abort
(void)
{
	gamma_abort_(my_rnd_key);
	my_par_pid = 0;
	gamma_resume_irq();
	return 0;
}


int fgamma_abort__ (int *res)
{
	*res = gamma_abort();
	return 0;
}


/***************************************************************************/
/*		Memory management routines.				   */
/***************************************************************************/


static volatile char defeat_cc_opt;

static inline int
gamma_mlock_lib
(void *buffer, unsigned long len,int flag)
{
	/*
	 *  Lock/Unlock a contiguous region of virtual memory into RAM.
	 *  This is good for both data structures (flag=1) and
	 *  receiver handlers (flag=0).
	 *  If flag=1, pre-fetch the memory buffer.
	 */
	char *p,*end;
	long res;
	if (buffer == NULL) return 0;
	p = (char *)buffer;
	end = (char *)(((unsigned long)(p+len-1+PAGE_SIZE))&~(PAGE_SIZE-1));

	/*  Lock buffer into RAM  */
	TRAP_P2(GAMMA_MLOCK,res,buffer,len);
	if (res) {
		fprintf(stderr, "gamma_mlock_lib(): could not lock.  ERR=%ld\n",
			res);
		return -GAMMA_ERR_GENERIC;
	}
	if (flag)
		/*
		 *  Buffer is for data:
		 *  Pre-fetch its logical pages by touching (on write)
		 *  one byte per page.
		 *  Must defeat compiler optimizations here...
		 */
		for (;p<end;p+=PAGE_SIZE) {
			defeat_cc_opt = *p;
			*p = defeat_cc_opt;
		}
	else
		/*
		 *  Buffer is for code:
		 *  Pre-fetch its logical pages by touching (on read)
		 *  one byte per page.
		 *  Must defeat compiler optimizations here...
		 */
		for (;p<end;p+=PAGE_SIZE)
			defeat_cc_opt += *p;
	return 0;
}


int gamma_mlock(void *buffer, unsigned long len)
{
	/*
	 *  Lock a data structure spanning a contiguous region.
	 *  This must be used in advance if the data structure is to
	 *  receive data through a GAMMA port.
	 */
	return gamma_mlock_lib(buffer,len,1);
}


int fgamma_mlock__(void *buffer, unsigned long *len, int *res)
{
	*res = gamma_mlock_lib(buffer,*len,1);
	return 0;
}


int gamma_mlock_funct(void *f, unsigned long len)
{
	/*
	 *  Lock in RAM function "f", referenced by generic pointer.
	 *  Will not pre-fetch (it is forbidden).
	 *  This must be used in advance if the function is to be
	 *  invoked by a GAMMA handler.
	 */
	return gamma_mlock_lib(f,len,0);
}


int fgamma_mlock_funct__(void *f, unsigned long *len, int *res)
{
	*res = gamma_mlock_funct(f,*len);
	return 0;
}


int gamma_munlock(void *buffer, unsigned long len)
{
	long res;

	/*
	 *  Unlock anything previously locked by gamma_mlock() or
	 *  gamma_mlock_funct().
	 */
/*
        __asm__ __volatile__ (
		"int $0x80"
		:
		"=a"(res)
		:
		"0"(LINUX_MUNLOCK),
		"b"((long)buffer),
		"c"((long)len)
	);
*/
	TRAP_P2(GAMMA_MUNLOCK,res,buffer,len);
	if (res) {
		fprintf(stderr,"gamma_munlock(): could not unlock.  ERR=%ld\n",
			res);
		return -GAMMA_ERR_GENERIC;
	}

	return 0;
}


int fgamma_munlock__(void *buffer, unsigned long *len, int *res)
{
	*res = gamma_munlock(buffer,*len);
	return 0;
}


int gamma_munlockall(void)
{
	/*  Unlock anything previously locked.  */
	long res;

	TRAP_P0(GAMMA_MUNLOCKALL,res);
	if (res) {
		fprintf(stderr,"gamma_munlockall(): could not unlock.  ERR=%ld\n",
			res);
		return -GAMMA_ERR_GENERIC;
	}
	return 0;
}


int fgamma_munlockall__(int *res)
{
	*res = gamma_munlockall();
	return 0;
}


/***************************************************************************/
/*		Miscellaneous routines.					   */
/***************************************************************************/


int gamma_atomic(void (*funct)())
{
	/*
	 *  Executes function "funct" in safe (wrt rec handlers) mode,
	 *  as it were invoked as a GAMMA handler.
	 */
	gamma_spinlock_take();
	funct();
	gamma_spinlock_release();
	return 0;
}


int fgamma_atomic__ (void (*funct)(), int *res)
{
	*res = gamma_atomic(funct);
	return 0;
}


int gamma_how_many_nodes(void)
{
        return how_many_inst;
}


int fgamma_how_many_nodes__ (int *res)
{
        *res = how_many_inst;
	return 0;
}


int gamma_my_par_pid(void)
{
        return my_par_pid;
}


int fgamma_my_par_pid__ (int *res)
{
        *res = my_par_pid;
	return 0;
}


int gamma_my_node(void)
{
        return my_instnum;
}


int fgamma_my_node__ (int *res)
{
	*res = my_instnum;
        return 0;
}


int gamma_name2i(char *node_name)
{
	int i;
	for (i=0; i<how_many_inst; i++)
		if (!strcmp(i2name_table[i],node_name)) return i;
	return -GAMMA_ERR_GENERIC;
}


/***************************************************************************/
/*		Time measurement functions, for performance measurements.  */
/*              To be used, they require the symbol CLOCK (gamma_time.h)   */
/*		to be set to the frequency of the local workstation in MHz.*/
/***************************************************************************/


int gamma_timers(time_586 from_t, time_586 to_t)
{
	/*
	 *  Force printout of time measurements taken at kernel level.
	 */
	long res;
	TRAP_P4(GAMMA_TIMERS,res,from_t.lo,from_t.hi,to_t.lo,to_t.hi);
	return (int)res;
}


int fgamma_time__ (unsigned long *hi,unsigned long *lo)
{
	/*
	 *  Read and store the current value of the 64-bit rdtsc register
	 *  of Pentium CPU.
	 */
	time_586 x;
	gamma_time(x);
	*hi = x.hi;
	*lo = x.lo;
	return 0;
}


int fgamma_time_diff__ (
	unsigned long *bhi,unsigned long *blo,
	unsigned long *ahi,unsigned long *alo,
	double *time_diff
)
{
	/*
	 *  Computes the time difference between (bhi,blo) and (ahi,alo)
	 *  in microsecond, where (bhi,blo) and (ahi,alo) are time instants
	 *  as previously read by the rdtsc register (function gamma_time()).
	 */
	time_586 b,a;
	b.hi=*bhi;
	b.lo=*blo;
	a.hi=*ahi;
	a.lo=*alo;
	*time_diff=gamma_time_diff(b,a);
	return 0;
}


/***************************************************************************/
/*		Collective communication routines.			   */
/***************************************************************************/


static int gamma_sync_safe(void)
{
	/*
	 *  A special barrier synchronization
	 *  used only at gamma_init() time.
	 */
#ifndef SYNC_SAFE_BCAST_SW
	if (!my_instnum) {
		if (how_many_inst == 1) return 0;
		if (!gamma_wait(0,how_many_inst-1))
			if (!gamma_SEND(0,NULL,0)) return 0;
	} else {
		if (!gamma_SEND(0,NULL,0))
			if (!gamma_wait(0,1)) return 0;
	}
	fprintf(stderr,"gamma_sync_safe():  Failure.\n");
	gamma_abort();
	exit(1);
#else
	int i;
	if (!my_instnum) {
		if (how_many_inst == 1) return 0;
		if (!gamma_wait(gamma_physical_nodes,how_many_inst - 1)) {
			for (i=1; i<how_many_inst; i++)
				if (gamma_SEND(i,NULL,0)) goto exit_failure;
			return 0;
		}
	}
	if (!gamma_SEND(0,NULL,0))
       		if (!gamma_wait(0,1)) return 0;

exit_failure:

	fprintf(stderr,"gamma_sync_safe():  Failure.\n");
	gamma_abort();
	exit(1);
#endif
}


int gamma_sync(void)
{
#ifdef USE_FAST_BARRIER_HALFDUPLEX
/*
 *  Barrier synchronization, using the optimal algorithm for half-duplex
 *  repeater hubs.  See
 *  Chiola, Ciaccio, "Fast Barrier Synchronization on Shared Fast Ethernet".
 *  LNCS 1362, Springer.
 */
	if (how_many_inst == 1) return 0;
	if (!my_instnum) {
		/*  Master instance.  */
		/*
		 *  Here waiting for SYNC_SUBSETS msgs would be wrong
		 *  when how_many_inst < SYNC_SUBSETS.
		 */
		if (gamma_wait(SYNC_PORT,
			(how_many_inst-1) % proc_chain_len ?
				(how_many_inst-1)/proc_chain_len + 1 :
				(how_many_inst-1)/proc_chain_len)
		)
			return -GAMMA_ERR_GENERIC;
		if (gamma_SEND(SYNC_PORT,NULL,0) < 0)
			return -GAMMA_ERR_GENERIC;
		return 0;
	}
	/*  Slave instances.  */
	if (	(proc_chain_len!=1) &&
		(my_instnum % proc_chain_len) &&
		(my_instnum != (how_many_inst-1))
	)
		/*
		 *  Only slaves which are not last in their respective
		 *  chains must wait for "enter barrier" msgs from
		 *  their successors.
		 */
		if (gamma_wait(SYNC_PORT,1))
			return -GAMMA_ERR_GENERIC;
	if (gamma_SEND(SYNC_PORT,NULL,0) < 0)
		return -GAMMA_ERR_GENERIC;
	if (gamma_wait(SYNC_PORT,1))
		return -GAMMA_ERR_GENERIC;
	return 0;
#else
/*
 *  Barrier synchronization, using a binary tree algorithm.
 *  Suitable to full-duplex Ethernet switches.
 */
	int i;

	for (i=1;i<MAX_NUM_NODES*MAX_NUM_LOCAL_INST;i<<=1) {
		if (my_instnum & i)
			break;
		if ((my_instnum+i) >= how_many_inst)
			break;
		if (gamma_wait(SYNC_PORT,1))
			return -GAMMA_ERR_GENERIC;
	}
	if (gamma_SEND(SYNC_PORT,NULL,0) < 0)
		return -GAMMA_ERR_GENERIC;
	if (my_instnum)
		if (gamma_wait(SYNC_PORT,1))
			return -GAMMA_ERR_GENERIC;
	return 0;
#endif
}


int fgamma_sync__ (int *res)
{
	*res = gamma_sync();
	return 0;
}


int gamma_bcast(void *data, unsigned long len, unsigned long root)
{
	/*
	 *  Broadcast.
	 *  WARNING:  the data buffers must have previously been locked in RAM
	 *  (gamma_mlock()), or at least must have been "write touched" so that
	 *  likely they are mapped in RAM.
	 */
	if ((unsigned long)my_instnum != root) {
		gamma_post_recv(BCAST_PORT,data,len);
/*
 *  This is necessary even with pure polling recv, because GAMMA uses signals
 *  which can act as IRQs sometimes.
 */
		gamma_sync();

		if (gamma_wait(BCAST_PORT,1))
			return -GAMMA_ERR_GENERIC;
		return 0;
	}
/*
 *  This is necessary even with pure polling recv, because GAMMA uses signals
 *  which can act as IRQs sometimes.
 */
	gamma_sync();

	if (gamma_send_flowctl(BCAST_PORT,data,len) < 0)
		return -GAMMA_ERR_GENERIC;
	return 0;
}


int fgamma_bcast__ (
	void *data, unsigned long *len, unsigned long *root, int *res
)
{
	*res = gamma_bcast(data, *len, *root);
	return 0;
}


/***************************************************************************/
/*		Other miscellaneous functions.				   */
/***************************************************************************/


int
gamma_vm_abort
(unsigned int par_pid)
{
	long res;

	do TRAP_P2(GAMMA_VM_ABORT,res,par_pid,0);
	while (res == -GAMMA_ERR_TRYAGAIN);
	return 0;
}


static int gamma_register(u_int32_t rnd_key)
{
	long res;
	TRAP_P1(GAMMA_REGISTER,res,rnd_key);
	return (int)res;
}


static int gamma_register_nospawn(u_int32_t rnd_key)
{
	long res;
	TRAP_P1(GAMMA_REGISTER_NOSPAWN,res,rnd_key);
	return (int)res;
}


static int getconfig_scanline(FILE *f, int node)
{
#ifndef OLDSTYLE
	int i;
	char *s, *next;
	char tmp_eth_string[64];
	eth_addr e;
#endif


#ifdef OLDSTYLE
#ifdef MORE_NICS_SAMENODE_ROUTING
	if ( fscanf(f,"%s %i %i %i %i %i %i %hd",
		host_table[node],
		(int *)&(eth[node].addr[0]),
		(int *)&(eth[node].addr[1]),
		(int *)&(eth[node].addr[2]),
		(int *)&(eth[node].addr[3]),
		(int *)&(eth[node].addr[4]),
		(int *)&(eth[node].addr[5]),
		&(nic[node])) == EOF
	)
#else
	if ( fscanf(f,"%s %i %i %i %i %i %i",
		host_table[node],
		(int *)&(eth[node].addr[0]),
		(int *)&(eth[node].addr[1]),
		(int *)&(eth[node].addr[2]),
		(int *)&(eth[node].addr[3]),
		(int *)&(eth[node].addr[4]),
		(int *)&(eth[node].addr[5])) == EOF
	)
#endif
		return 0;

	eth[node].addrl =
		((unsigned long)eth[node].addr[0]) +
		((unsigned long)eth[node].addr[1] << 8) +
		((unsigned long)eth[node].addr[2] << 16) +
		((unsigned long)eth[node].addr[3] << 24);
	
	eth[node].addrw =
		((unsigned long)eth[node].addr[4]) +
		((unsigned long)eth[node].addr[5] << 8);
#else

#ifndef MORE_NICS_SAMENODE_ROUTING
	if (fscanf(f, "%s %s", host_table[node], tmp_eth_string) != 2)
#else
	if (fscanf(f, "%s %s %hd\n", host_table[node], tmp_eth_string, &(nic[node])) != 3)
#endif
		return 0;

	s = tmp_eth_string;
	for (i=0; i<6; i++)
	{
		e.addr[i] = strtol(s, &next, 16);
		if(*next == ':')
			s = next+1;
		else break;
	}
  
	if (*next != '\0')
		return 0;

	e.addrl =
		((unsigned long)e.addr[0]) +
		((unsigned long)e.addr[1] << 8) +
		((unsigned long)e.addr[2] << 16) +
		((unsigned long)e.addr[3] << 24);
	
	e.addrw =
		((unsigned long)e.addr[4]) +
		((unsigned long)e.addr[5] << 8);

	eth[node] = e;
#endif

	return 1;
}


int gamma_getconfig(int flag)
{

	/*
	 *  Read the host names and the MAC addresses of the stations
	 *  from file /etc/gamma.conf,
	 *  then register the MAC addresses into the kernel.
	 *  Fills the table host_table[] of the workstation names.
	 *  Set the value for gamma_physical_nodes and gamma_phys_whoami.
	 */
	long res;
	int num_nodes;
	FILE *f;


	/*  read the configuration file of GAMMA: /etc/gamma.conf  */

	f = fopen("/etc/gamma.conf","r");
	if (f == NULL) {
fprintf(stderr,"gamma_getconfig():  Could not read /etc/gamma.conf\n");
		return -GAMMA_ERR_GENERIC;
	}

	num_nodes = 0;

	while (getconfig_scanline(f,num_nodes))
		if (++num_nodes >= MAX_NUM_NODES)
			break;
	fclose(f);

	if (num_nodes == 0) {
fprintf(stderr,"gamma_getconfig():  empty /etc/gamma.conf file\n");
		return -GAMMA_ERR_GENERIC;
	}

	gamma_physical_nodes = num_nodes;

#ifdef MORE_NICS_SAMENODE_ROUTING
	TRAP_P4(GAMMA_GETCONFIG,res,num_nodes,eth,nic,flag);
#else
	TRAP_P4(GAMMA_GETCONFIG,res,num_nodes,eth,NULL,flag);
#endif

	if (res >= 0)
		gamma_phys_whoami = res;

	return (int)res;

}

u_int32_t gamma_getkey
(void)
{
	u_int32_t res;
	TRAP_P0(GAMMA_GETKEY,res);
	return res;
}


void gamma_busywait_usec
(unsigned int n)
{
	time_586 a,b;
	gamma_time(a);
	for (;;) {
		gamma_time(b);
		if ((unsigned int)gamma_time_diff(b,a) > n) break;
	}
}


int gamma_printk
(long val)
{
	long res;
	TRAP_P1(GAMMA_PRINTK,res,val);
	return (int)res;
}

int fgamma_printk__(long *val, int *res)
{
	*res = gamma_printk(*val);
	return 0;
}


int gamma_setcredit
(unsigned int size)
{
	long res;
	TRAP_P1(GAMMA_SETCREDIT,res,size);
	return (int)res;
}


int gamma_resetcredit
(void)
{
	long res;
	TRAP_P0(GAMMA_RESETCREDIT,res);
	return (int)res;
}


int gamma_getcredit
(void)
{
	long res;
	TRAP_P0(GAMMA_GETCREDIT,res);
	return (int)res;
}


int gamma_setmaxpolls
(unsigned int n)
{
	long res;
	TRAP_P1(GAMMA_SETMAXPOLLS,res,n);
	return (int)res;
}


int gamma_resetmaxpolls
(void)
{
	long res;
	TRAP_P0(GAMMA_RESETMAXPOLLS,res);
	return (int)res;
}


int gamma_getmaxpolls
(void)
{
	long res;
	TRAP_P0(GAMMA_GETMAXPOLLS,res);
	return (int)res;
}
