/***************************************************************************
	Routines for supporting the launch of a GAMMA job on the cluster.
	Read COPYRIGHT file in this directory
****************************************************************************/

/*
 *  Try to create a virtual GAMMA across the network, assigning it a unique
 *  par_pid.
 *  On success it returns the par_pid, and a list of indexes of the nodes
 *  involved in the virtual GAMMA as its 2th arg.
 *  NOTE:  the arguments "user_space_node_list_***" point to user space!!!
 */
asmlinkage long
sys_gamma_vm_create
(
	u_int32_t rnd_key,
	u_int32_t num_instances,
	void *user_space_node_list_in,
	void *user_space_node_list_out
)
{

	int i,j,par_pid,found,procnum;
	par_entry *p;
	instance_entry *inst;
	port_entry *pp;


	if (num_instances < 1
	 || num_instances > total_num_nodes*MAX_NUM_LOCAL_INST) {
printg("%s: wrong or unfeasible no. of instances\n",__FUNCTION__);
		return -GAMMA_ERR_GENERIC;
	}

/*
 * choose a fresh par_pid, if one is available.
 * par_pid #0 is reserved!
 * Mutex mandatory on par_proc_table[], reuse gamma_tx_lock for this.
 */
	spin_lock(&gamma_tx_lock);
	found = 0;
	for (par_pid=1; par_pid <= MAX_PAR_PID; par_pid++) {
		if ((par_pid-1-phys_whoami) % total_num_nodes)
			continue;
		if (par_proc_table[par_pid] != NULL)
			continue;
		found = 1;
		break;
	}

	if (!found) {
printg("sys_gamma_vm_create(): could not find a par_pid\n");
		spin_unlock(&gamma_tx_lock);
		return -GAMMA_ERR_GENERIC;
	}

/*  Allocate a descriptor for the new virtual GAMMA  */
	par_proc_table[par_pid] = (par_entry *)
		__get_free_pages(GFP_KERNEL,par_entry_magnitude);

/*  Now the entry in par_proc_table[] has been taken...  */
	spin_unlock(&gamma_tx_lock);

	p = par_proc_table[par_pid];
	if (p == NULL) {
printg("sys_gamma_vm_create(): kmalloc() failed\n");
		return -GAMMA_ERR_GENERIC;
	}

	memset(p,0,sizeof(par_entry));

/*  No process instances yet  */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		inst = &p->inst_table[i];
		inst->pid = -1;
		inst->task = NULL;
		inst->whoami = -1;
	}

	p->vm.random_key = rnd_key;
	p->vm.num_instances = num_instances;

	procnum = 0;

	if (user_space_node_list_in == NULL) {

/*
 *  Choose "num_instances" nodes to create a virtual GAMMA,
 *  indexing them from 0 on, fill in the virtual GAMMA
 *  descriptor, find out how many process instances are allocated on the
 *  local node ("procnum"), and allocate room for pre-computed pkt headers.
 *  Nodes are chosen circularly in file /etc/gamma.conf, starting
 *  from local node.
 */

		for (i=0; i<num_instances; i++) {
			p->vm.node_list[i] = (phys_whoami+i) % total_num_nodes;
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create():  too many proc instances on this node.  Aborting...\n");
				par_proc_table[par_pid] = NULL;
				free_pages(
					(unsigned long)p,par_entry_magnitude);
				return -GAMMA_ERR_GENERIC;
			}
			inst = &p->inst_table[procnum-1];
			inst->whoami = i;
			for (j=0; j<=MAX_PORT_NUMBER; j++) {
				pp = &inst->port_table[j];
				pp->task = NULL;
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_intra_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = kmalloc(
						HEADER_SIZE+MAX_1ST_PIECE_SIZE,
						GFP_KERNEL);
				if (pp->header == NULL) {
printg("sys_gamma_vm_create():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
			}
		}

	} else {

/*
 *  Take "num_instances" nodes from the user-provided list of machines,
 *  then build the virtual GAMMA similar to the above "if" branch.
 */
		gamma_copy_from_user(
			p->vm.node_list,user_space_node_list_in,
			num_instances*sizeof(unsigned short)
		);

		for (i=0; i<num_instances; i++) {
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("sys_gamma_vm_create():  too many proc instances on this node.  Aborting...\n");
				par_proc_table[par_pid] = NULL;
				free_pages(
					(unsigned long)p,par_entry_magnitude);
				return -GAMMA_ERR_GENERIC;
			}
			inst = &p->inst_table[procnum-1];
			inst->whoami = i;
			for (j=0; j<=MAX_PORT_NUMBER; j++) {
				pp = &inst->port_table[j];
				pp->task = NULL;
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_intra_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = kmalloc(
						HEADER_SIZE+MAX_1ST_PIECE_SIZE,
						GFP_KERNEL);
				if (pp->header == NULL) {
printg("sys_gamma_vm_create():  kmalloc() failed\n");
					return -GAMMA_ERR_GENERIC;
				}
			}
			}
		}

	}

/*  The following actually marks the end of virtual GAMMA initialization.  */
	p->num_locinst = procnum;

/*  Return the chosen nodes to the caller in user space  */
	gamma_copy_to_user(
		user_space_node_list_out,p->vm.node_list,
		num_instances*sizeof(unsigned short)
	);

/*  Register the virtual GAMMA across the network, via broadcast  */
/*  Some switches (3COM?) may need this, as they tend to sleep sometimes...  */
#if 0
	for (j=3;j;j--)
#endif
		gamma_kernel_send(
			BROADCAST,par_pid,VM_CREATE,
			(void *)(&(p->vm)),sizeof(vm_descriptor),1
		);

	return par_pid;

}


asmlinkage long
sys_gamma_vm_create_bykey
(
	u_int32_t rnd_key,
	u_int32_t num_instances,
	void *user_space_node_list
)
{

	int i,j,k,par_pid,procnum,found;
	par_entry *p;
	instance_entry *inst;
	port_entry *pp;


	if (num_instances < 1
		|| num_instances > total_num_nodes*MAX_NUM_LOCAL_INST) {
printg("%s: wrong or unfeasible no. of instances\n",__FUNCTION__);
		return -GAMMA_ERR_GENERIC;
	}

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		p = par_proc_table[par_pid];
		if (p == NULL) continue;
		if (p->vm.random_key == rnd_key)
			break;
	}

/*  A virtual GAMMA should already exist.  */
	if (par_pid > MAX_PAR_PID) {
printg("%s: key %u not created\n",__FUNCTION__,rnd_key);
		return -GAMMA_ERR_GENERIC;
	}

/*  We should have collected "num_instances" participants already.  */
	if (p->vm.num_instances != num_instances) {
printg("%s: num_instances mismatch\n",__FUNCTION__);
		return -GAMMA_ERR_GENERIC;
	}

/*  Always give instance #0 to creator (master)  */
	p->inst_table[0].pid = current->pid;
	p->inst_table[0].whoami = 0;

/*  Build list of involved physical nodes, by scanning the wish list.  */
	p->vm.node_list[0] = phys_whoami;
	i = 1;
	found = 0;
	for (j=0; j<MAX_NUM_NODES; j++) {
		for (k=0; k<MAX_NUM_LOCAL_INST; k++) {
			if (p->wish_pid[j*MAX_NUM_LOCAL_INST+k] < 0)
				continue;
			p->vm.node_list[i++] = j;
			if (i >= num_instances) {
				found = 1;
				break;
			}
		}
		if (found) break;
	}
	if (!found) {
printg("%s: should not have arrived here\n",__FUNCTION__);
		return -GAMMA_ERR_GENERIC;
	}

/*  Init the instances that are local to this physical node.  */
	procnum = 0;
	for (i=0; i<num_instances; i++) {
		if (p->vm.node_list[i] != phys_whoami) continue;
		procnum++;
		if (procnum > MAX_NUM_LOCAL_INST) {
printg("%s:  too many proc instances on this node.  Aborting...\n",
__FUNCTION__);
			par_proc_table[par_pid] = NULL;
			free_pages((unsigned long)p,par_entry_magnitude);
			return -GAMMA_ERR_GENERIC;
		}
		inst = &p->inst_table[procnum-1];
		inst->whoami = i;
		for (j=0; j<=MAX_PORT_NUMBER; j++) {
			pp = &inst->port_table[j];
			pp->task = NULL;
			pp->semaphore =
			pp->boundary_1p =
			pp->in_size =
			pp->in_room = 0;
			pp->in_buffer =
			pp->in_progress = NULL;
			pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
			pp->partner_is_intra_node =
			pp->bcast_was_susp = 0;
#endif
			pp->header = kmalloc(
					HEADER_SIZE+MAX_1ST_PIECE_SIZE,
					GFP_KERNEL);
			if (pp->header == NULL) {
printg("sys_gamma_vm_create_bykey():  kmalloc() failed\n");
				return -GAMMA_ERR_GENERIC;
			}
		}
	}
/*  The following actually marks the end of virtual GAMMA initialization.  */
	p->num_locinst = procnum;

/*  Return the node list to the caller in user space.  */
	gamma_copy_to_user(
		user_space_node_list,p->vm.node_list,
		num_instances*sizeof(unsigned short)
	);

/*  Register the virtual GAMMA across the network, via broadcast  */
/*  Some switches (3COM?) may need this, as they tend to sleep sometimes...  */
#if 0
	for (j=3;j;j--)
#endif
		gamma_kernel_send(
			BROADCAST,par_pid,VM_CREATE,
			(void *)(&(p->vm)),sizeof(vm_descriptor),1
		);

	return par_pid;

}


/*
 *  The invoking process informs all nodes that it would need a
 *  virtual GAMMA identified by "random_key" and expects "num_instances"
 *  other processes to join the virtual GAMMA as well.
 */
asmlinkage long
sys_gamma_vm_wish
(
	u_int32_t rnd_key,
	u_int32_t num_instances,
	int is_master,
	int first_time
)
{

	int		i,par_pid,found;
	par_entry	*p;
	instance_entry	*inst;
	wish_t		wish;


	if (num_instances < 1
	|| num_instances > total_num_nodes*MAX_NUM_LOCAL_INST)
		return -GAMMA_ERR_GENERIC;

/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
/* NOTE: "0" is purely arbitrary here, and possibly wrong, but harmless */
	sys_gamma_poll(0,0);

	if (is_master && first_time) {

/*
 * choose a par_pid, if one is available.
 * par_pid #0 is reserved!
 * Mutex mandatory on par_proc_table[], reuse gamma_tx_lock for this.
 */
		spin_lock(&gamma_tx_lock);
		found = 0;
		for (par_pid=1; par_pid <= MAX_PAR_PID; par_pid++) {
			if ((par_pid-1-phys_whoami) % total_num_nodes)
				continue;
			if (par_proc_table[par_pid] != NULL)
				continue;
			found = 1;
			break;
		}
		if (!found) {
printg("sys_gamma_vm_wish(): master: could not find a par_pid\n");
			spin_unlock(&gamma_tx_lock);
			return -GAMMA_ERR_GENERIC;
		}

/*  Allocate a descriptor for the new virtual GAMMA  */
		p = (par_entry *)
			__get_free_pages(GFP_KERNEL,par_entry_magnitude);
		if (p == NULL) {
printg("sys_gamma_vm_wish(): kmalloc() failed\n");
			spin_unlock(&gamma_tx_lock);
			return -GAMMA_ERR_GENERIC;
		}
		memset(p,0,sizeof(par_entry));
		p->vm.random_key = rnd_key;

/*  make all slots for local instances empty  */
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			inst = &p->inst_table[i];
			inst->pid = -1;
			inst->task = NULL;
			inst->whoami = -1;
		}
/*  nobody has joined the party yet  */
		for (i=0; i<MAX_NUM_NODES*MAX_NUM_LOCAL_INST; i++)
			p->wish_pid[i] = -1;

/*  for now, just the caller process/node has joined the party  */
		p->vm.num_instances = 1;
		p->inst_table[0].pid = current->pid;
		p->inst_table[0].task = current;

/*  make the descriptor available so others can join  */
		par_proc_table[par_pid] = p;

/*  Now the entry in par_proc_table[] has been taken...  */
		spin_unlock(&gamma_tx_lock);

/*  come later to see if slaves have joined the party  */
		return -GAMMA_ERR_TRYAGAIN;
	}

	if (is_master && !first_time) {
/*  see if slaves have joined the party  */

/*  find the virtual GAMMA by key  */
		p = NULL;
		found = 0;
		for (par_pid=1; par_pid <= MAX_PAR_PID; par_pid++) {
			if ((par_pid-1-phys_whoami) % total_num_nodes)
				continue;
			p = par_proc_table[par_pid];
			if (p == NULL) continue;
			if (p->vm.random_key != rnd_key) continue;
			found = 1;
			break;
		}
		if (!found) {
printg("sys_gamma_vm_wish(): master: could not find entry for key %u\n",
rnd_key);
			return -GAMMA_ERR_GENERIC;
		}

		if (p->vm.num_instances < num_instances)
			return -GAMMA_ERR_TRYAGAIN;

/*  enough nodes have joined, so quit successfully  */
		return 0;
	}

/*  here slaves only  */
/*
 *  express a wish to join the virtual GAMMA identified by "rnd_key".
 *  Repeatedly tell that to every machine in the cluster.
 *  This way, master will eventually catch the wish.
 */

/*  find the virtual GAMMA by key, in case it is available on local node  */
	p = NULL;
	found = 0;
	for (par_pid=1; par_pid <= MAX_PAR_PID; par_pid++) {
		if ((par_pid-1-phys_whoami) % total_num_nodes) continue;
		p = par_proc_table[par_pid];
		if (p == NULL) continue;
		if (p->vm.random_key != rnd_key) continue;
		found = 1;
		break;
	}

	if (found) {
/*  Express a wish to join the virtual GAMMA found locally.  */
/*  Mutex mandatory on p->wish_pid[], reuse gamma_tx_lock for this.  */
		spin_lock(&gamma_tx_lock);
		for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
			if (p->wish_pid[phys_whoami*MAX_NUM_LOCAL_INST+i]
			== current->pid) {
				/* wish already present */
				spin_unlock(&gamma_tx_lock);
				return -GAMMA_ERR_TRYAGAIN;
			}
			if (p->wish_pid[phys_whoami*MAX_NUM_LOCAL_INST+i]
			!= -1)
				continue;

DEBUG_L(printg(
"%s: slave: put a wish on locally found virtual GAMMA (key %u)\n",
__FUNCTION__,rnd_key
));

			p->vm.num_instances++;
			p->wish_pid[phys_whoami*MAX_NUM_LOCAL_INST+i]
				= current->pid;
			spin_unlock(&gamma_tx_lock);
			return -GAMMA_ERR_TRYAGAIN;
		}

printg(
"%s: reject wish from pid %u for virtual GAMMA key %u: no free slot\n",
__FUNCTION__,current->pid,rnd_key
);

		spin_unlock(&gamma_tx_lock);
		return -GAMMA_ERR_GENERIC;
	}

/*  Express a wish to join the virtual GAMMA being create somewhere remote  */
	wish.key = rnd_key;
	wish.pid = current->pid;
	gamma_kernel_send(
		BROADCAST,DONT_MIND,VM_WISH,&wish,sizeof(wish),1);
	return -GAMMA_ERR_TRYAGAIN;

}


/*
 *  The invoking process needs the mapping "instance # --> physical node #" .
 *  This functionality is also contained in sys_gamma_vm_create() and
 *  sys_gamma_vm_create_bykey().
 *  NOTE:  the argument "user_space_node_list" points to user space!!!
 */
asmlinkage long
sys_gamma_vm_getinfo
(u_int32_t par_pid, void *user_space_node_list)
{

/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	if (par_proc_table[par_pid] == NULL)
		return -GAMMA_ERR_GENERIC;

	gamma_copy_to_user(
	  user_space_node_list,par_proc_table[par_pid]->vm.node_list,
	  par_proc_table[par_pid]->vm.num_instances*sizeof(unsigned short)
	);

	return 0;

}


/*
 * Read in the config file, possibly clean some data struct
 * returns the physical rank of this machine in the cluster.
 */
asmlinkage long
sys_gamma_getconfig
(
	u_int32_t n,
	void *user_space_eth_table,
	void *user_space_routing_table, /* != NULL iff #defined MORE_NICS... */
	int flag
)
{

	int i,j,k,n_old;
	long res;


/*  Check if GAMMA device(s) are present and ifconfig'ed  */
	if (!num_gamma_devices) {

printg("sys_gamma_getconfig(): no GAMMA devices found.\n");

		return -GAMMA_ERR_NODEV;
	}

	for (k=0,i=0; i<num_gamma_devices; i++)
		if (netif_running(gamma_devices[i]))
			k++;

	if (k != num_gamma_devices) {

printg(
"sys_gamma_getconfig(): found a not configured NIC.  Please ifconfig!\n"
);

		return -GAMMA_ERR_NOIFCONFIG;
	}

	if (!flag && !gamma_initialized) {

printg(
"sys_gamma_getconfig(): attempting to use GAMMA before initialization!\n"
);

		return -GAMMA_ERR_NOINIT;
	}

	if (flag) {

/*  Init and clean a lot of thing ;-)  */

		if (n < 1 || n > MAX_NUM_NODES) {

printg("sys_gamma_getconfig(): wrong or unfeasible number of nodes\n");

			return -GAMMA_ERR_GENERIC;
		}

/*  Mutex is safer; re-use gamma_tx_lock for this  */
		spin_lock(&gamma_tx_lock);

		if (total_num_nodes)
			n_old = total_num_nodes;
		else
/*
 *  This might be first invocation after boot (that is, after gamma_start()).
 *  Thus, we have allocated MAX_NUM_NODES Tx tmp rings (presumably more than
 *  needed).
 */
			n_old = MAX_NUM_NODES;
/*  De-allocate Tx tmp rings.  */
		for (i=0; i<n_old; i++)
			for (j=0; j<GAMMA_TX_WIN_SIZE; j++)
				if (tx_tmp_ring[i].buf[j].b != NULL) {
					kfree(tx_tmp_ring[i].buf[j].b);
					tx_tmp_ring[i].buf[j].b = NULL;
				}

		total_num_nodes = n;

/*
 *  Set the initial credit towards each destination.
 *  It may be safer to leave two (or more?) available Rx slots
 *  off-credit for service (non-user) communications.
 */
		for (i=0; i<total_num_nodes; i++) {
			initial_credit[i] = credit_size;
			credit_ceil[i] = credit_size;
			credit_growth_cnt[i] = 0;
			atomic_set(&local_credit[i],initial_credit[i]);
		}

/*  Allocate and clean the Tx tmp rings.  */

		for (i=0; i<total_num_nodes; i++) {
			for (j=0; j<GAMMA_TX_WIN_SIZE; j++) {
				tx_tmp_ring[i].buf[j].b = (unsigned char *)
					kmalloc(FRAME_SIZ, GFP_KERNEL);
				if (tx_tmp_ring[i].buf[j].b == NULL) {
printg("sys_gamma_getconfig():  kmalloc() failed\n");
					spin_unlock(&gamma_tx_lock);
					return -GAMMA_ERR_GENERIC;
				}
				tx_tmp_ring[i].buf[j].l = 0;
			}
			tx_tmp_ring[i].tail = 0;
#ifdef SEQCHECK
/*  Reset the outgoing pkt numbering.  */
			tx_tmp_ring[i].tx_seq = 0;
#endif
		}
		for (j=0; j<GAMMA_TX_WIN_SIZE; j++)
			kernel_tx_tmp_ring.buf[j].l = 0;
		kernel_tx_tmp_ring.tail = 0;
#ifdef SEQCHECK
/*  Reset the outgoing pkt numbering.  */
		kernel_tx_tmp_ring.tx_seq = 0;

/*  Reset the tables of incoming pkt numbering and congestion flags.  */
		for (i=0; i<total_num_nodes; i++) {
			rx_seq[i] = rx_seq_bcast[i] = 0;
			rx_discard[i] = 0;
			congested[i] = 0;
		}
#endif

		gamma_must_yield.task = NULL;
		gamma_must_yield.time.hi = 0;
		gamma_must_yield.time.lo = 0;
		gamma_run_handler.task = NULL;

/*  get the table of MAC addresses from user space  */
		memset(eth_table,0,MAX_NUM_NODES*sizeof(eth_addr));
		gamma_copy_from_user(
			eth_table,user_space_eth_table,
			total_num_nodes*sizeof(eth_addr)
		);

#ifdef MORE_NICS_SAMENODE_ROUTING
/* copy the routing table from user space */
		gamma_copy_from_user(
			routing_table,user_space_routing_table,
			MAX_NUM_NODES*sizeof(unsigned short)
		);
#endif

		gamma_initialized = 1;

		spin_unlock(&gamma_tx_lock);

	}

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG(
	printg("sys_gamma_getconfig():  ROUTING_TABLE:\n");
	for (i=0;i<total_num_nodes;i++)
		printg("MAC: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x nic: %d\n",
		       eth_table[i].addr[0],
		       eth_table[i].addr[1],
		       eth_table[i].addr[2],
		       eth_table[i].addr[3],
		       eth_table[i].addr[4],
		       eth_table[i].addr[5],
		       routing_table[i]
		);
)
#else
DEBUG(
	printg("sys_gamma_getconfig():  ETH_TABLE:\n");
	for (i=0; i<total_num_nodes; i++)
               printg("%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		      eth_table[i].addr[0],
		      eth_table[i].addr[1],
		      eth_table[i].addr[2],
		      eth_table[i].addr[3],
		      eth_table[i].addr[4],
		      eth_table[i].addr[5]
	       );
)
#endif

/*  what node is this wrt the physical GAMMA?  */
	res = -1;
	for (i=0;i<total_num_nodes;i++) {
		for(k=0; k<num_gamma_devices; k++) {
			for (j=0;j<6;j++)
				if (eth_table[i].addr[j] != 
				    gamma_devices[k]->dev_addr[j])
					break;
			if (j==6) {
				res = i;
				goto end;
			}
		}
	}

 end:

	if (res < 0) {

printg("sys_gamma_getconfig(): local MAC addr not found in gamma.conf\n");

		return -GAMMA_ERR_NOGAMMACONF;
	}

DEBUG( printg("sys_gamma_getconfig(): phys_whoami = %ld\n", res) );
DEBUG( printg("sys_gamma_getconfig(): gamma_physical_nodes = %ld\n",
total_num_nodes) );

	phys_whoami = res;
	return res;

}


/*
 *  Registers the PID and the task pointer of the calling process
 *  with the virtual GAMMA specified by par_pid (which is supposed to exist).
 */
asmlinkage long
sys_gamma_register
(u_int32_t rnd_key)
{

	unsigned int i,j,par_pid,found;
	par_entry *p;
	instance_entry *inst;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t *shared_ring;
#endif


	found = 0;

DEBUG(printg(
"sys_gamma_register(): pid %d, key %u: entering\n",current->pid,rnd_key
));

	for (i=100;i;i--) {
		for (j=1; j<=MAX_PAR_PID; j++) {
			p = par_proc_table[j];
			if (p == NULL) continue;
			if (p->vm.random_key == rnd_key) {
				found = 1;
				break;
			}
		}
		if (found) break;
/*  NOTE: "0" is purely arbitrary here, and possibly wrong, but harmless  */
		sys_gamma_poll(0,0);  /* kernel_rec() eventll VM_CREATE */
		busywait(1000);
	}

/*  The virtual GAMMA must exist and be initialized before registering  */
	if (!found || ! p->num_locinst) {

printg("%s: key %u not created\n",__FUNCTION__,rnd_key);

		return -GAMMA_ERR_GENERIC;
	}

	par_pid = j;

DEBUG(
printg(
"sys_gamma_register(): virtual GAMMA %u (key %u):\nnum_instances %u\n",
	par_pid,rnd_key,p->vm.num_instances);
for (i=0; i<p->vm.num_instances; i++)
	printg("%u: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		i,
		eth_table[p->vm.node_list[i]].addr[0],
		eth_table[p->vm.node_list[i]].addr[1],
		eth_table[p->vm.node_list[i]].addr[2],
		eth_table[p->vm.node_list[i]].addr[3],
		eth_table[p->vm.node_list[i]].addr[4],
		eth_table[p->vm.node_list[i]].addr[5]
	);
);

/*
 *  Since more process instances are allowed on the same host, the
 *  caller has to find the right slot in the virtual GAMMA.
 *  Mutex is mandatory; re-use gamma_tx_lock for this.
 */
	spin_lock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	write_lock(&gamma_rxintranode_lock);
#endif
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		inst = &p->inst_table[i];
		if (inst->pid != -1 && inst->pid != current->pid)
/*  Another process already registered in this slot  */
			continue;
		if (inst->whoami >= 0) {
/*  A free, valid slot was found  */
			inst->pid = current->pid;
			inst->reenter = 0;
			inst->task = current;
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Init and allocate the ring of temp buffers
 *  for intra-node communication.
 */
			shared_ring = &inst->shared_ring;
			shared_ring->head = 0;
			atomic_set(&shared_ring->dirty, 0);
			atomic_set(&shared_ring->tail, 0);
			atomic_set(&shared_ring->room, SHARED_RING_SIZE);
			for (i=0; i < SHARED_RING_SIZE; i++) {
				shared_ring->desc[i].port = 0;
				shared_ring->desc[i].cnk_len = 0;
				shared_ring->desc[i].bitmap = IS_EMPTY;
				shared_ring->desc[i].buf =
					kmalloc(PAGE_SIZE,GFP_KERNEL);
				if (shared_ring->desc[i].buf == NULL) {
					printg("%s:  kmalloc() failed\n",
						__FUNCTION__);
					return -GAMMA_ERR_GENERIC;
				}
			}
#endif

DEBUG_L(printg(
"sys_gamma_register(): virtual GAMMA %u (key %u): whoami %d (pid %d)\n",
par_pid,rnd_key,inst->whoami,inst->pid
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
			write_unlock(&gamma_rxintranode_lock);
#endif
			spin_unlock(&gamma_tx_lock);
			return par_pid;
		}

DEBUG(printg(
"sys_gamma_register(): local node does not belong to virtual GAMMA %u\n",
par_pid
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		return -GAMMA_ERR_GENERIC;
	}

printg("sys_gamma_register(): par_pid %u: could not register\n",par_pid);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

}


asmlinkage long
sys_gamma_register_nospawn
(u_int32_t rnd_key)
{

	unsigned int i,j,par_pid,found;
	par_entry *p;
	instance_entry *inst;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t *shared_ring;
#endif


/*
 * this can be necessary to let more GAMMA jobs start simultaneously on the
 * same cluster
 */
	found = 0;
	for (i=100;i;i--) {
		for (j=1; j<=MAX_PAR_PID; j++) {
			p = par_proc_table[j];
			if (p == NULL) continue;
			if (p->vm.random_key == rnd_key) {
				found = 1;
				break;
			}
		}
		if (found) break;
/*  NOTE: "0" is purely arbitrary here, and possibly wrong, but harmless  */
		sys_gamma_poll(0,0);  /* kernel_rec() eventll VM_CREATE */
		busywait(1000);
	}

/*  The virtual GAMMA must exist and be initialized before registering  */
	if (!found || ! p->num_locinst) {

DEBUG(printg("%s: key %u not yet created\n",__FUNCTION__,rnd_key));

		return -GAMMA_ERR_TRYAGAIN;
	}

	par_pid = j;

{
DEBUG(
int i;
printg("%s: virtual GAMMA %d (key %u):\nnum_instances %u\n",__FUNCTION__,
par_pid,rnd_key,p->vm.num_instances);
for (i=0; i<p->vm.num_instances; i++)
	printg("%d: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
		i,
		eth_table[p->vm.node_list[i]].addr[0],
		eth_table[p->vm.node_list[i]].addr[1],
		eth_table[p->vm.node_list[i]].addr[2],
		eth_table[p->vm.node_list[i]].addr[3],
		eth_table[p->vm.node_list[i]].addr[4],
		eth_table[p->vm.node_list[i]].addr[5]
	);
);
}

/*
 *  Since more process instances are allowed on the same station, the
 *  caller has to find the right slot in the virtual GAMMA.
 *  Mutex is mandatory; re-use gamma_tx_lock for this.
 */
	spin_lock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	write_lock(&gamma_rxintranode_lock);
#endif
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		inst = &p->inst_table[i];
		if (inst->whoami < 0) {

DEBUG_L(printg(
"%s: local node does not belong to virtual GAMMA #%d\n",__FUNCTION__,par_pid
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
			write_unlock(&gamma_rxintranode_lock);
#endif
			spin_unlock(&gamma_tx_lock);
			return -GAMMA_ERR_GENERIC;
		}
		if (inst->pid != -1 && inst->pid != current->pid)
/*  Another process already registered in this slot  */
			continue;
/*  A free, valid slot was found  */
		inst->pid = current->pid;
		inst->reenter = 0;
		inst->task = current;
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Init and allocate the ring of temp buffers
 *  for intra-node communication.
 */
		shared_ring = &inst->shared_ring;
		shared_ring->head = 0;
		atomic_set(&shared_ring->dirty, 0);
		atomic_set(&shared_ring->tail, 0);
		atomic_set(&shared_ring->room, SHARED_RING_SIZE);
		for (i=0; i < SHARED_RING_SIZE; i++) {
			shared_ring->desc[i].port = 0;
			shared_ring->desc[i].cnk_len = 0;
			shared_ring->desc[i].bitmap = IS_EMPTY;
			shared_ring->desc[i].buf =
				kmalloc(PAGE_SIZE,GFP_KERNEL);
			if (shared_ring->desc[i].buf == NULL) {
				printg("%s:  kmalloc() failed\n",
					__FUNCTION__);
				return -GAMMA_ERR_GENERIC;
			}
		}
#endif

DEBUG_L(printg(
"%s: virtual GAMMA %d (key %u): whoami %d (pid %d)\n",__FUNCTION__,
par_pid,rnd_key,inst->whoami,inst->pid
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		return par_pid;
	}

printg("%s: par_pid %d:  could not register\n",__FUNCTION__,par_pid);

#ifdef MORE_INSTANCES_ON_SAME_NODE
	write_unlock(&gamma_rxintranode_lock);
#endif
	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

}


/*
 *  sys_gamma_init, parts 1 & 2 & 3.
 *  These GAMMA primitives look into par_proc_table for the caller's entry
 *  (which was assigned to the caller via a previous invocation to
 *  "gamma_register()")
 *  and then register the caller's tables of receiver buffers and semaphores.
 */

/*
 *  This routine returns the no. of nodes of the virtual GAMMA
 *  which the caller had previously registered to, and
 *  the instance no. of the caller.
 */
asmlinkage long
sys_gamma_init_part1
(
	void *user_space_msglen_ptr
)
{

	int found=0,i=0,par_pid;
	par_entry *p;
	instance_entry *inst;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		p = par_proc_table[par_pid];
		if (p == NULL) continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			inst = &p->inst_table[i];
			if (inst->pid == current->pid) {
				found = 1;
				break;
			}
		}
		if (found) break;
	}

	if (!found) return -GAMMA_ERR_GENERIC;

	inst->msglen_ptr = user_space_msglen_ptr;

        return (long)(
		((unsigned long)(p->vm.num_instances) << 16) +
		(unsigned long)p->inst_table[i].whoami
	);

}


/*
 *  This routine returns the "procnum" of the caller and the par_pid
 *  of the virtual GAMMA which the caller had previously registered to.
 */
asmlinkage long
sys_gamma_init_part2
(
	void *user_space_sem1_table_ptr,
	void *user_space_pre_sem_table_ptr
)
{

	int found=0,i=0,par_pid;
	par_entry *p;
	instance_entry *inst;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		p = par_proc_table[par_pid];
		if (p == NULL) continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			inst = &p->inst_table[i];
			if (inst->pid == current->pid) {
				found = 1;
				break;
			}
		}
		if (found) break;
	}

	if (!found) return -GAMMA_ERR_GENERIC;

/*  Register the tables of user-space semaphores.  */
#ifdef USE_DAMNICK
	inst->sem1_table = user_space_sem1_table_ptr;
	inst->pre_sem_table = user_space_pre_sem_table_ptr;
#endif
/*  Return the "procnum" of the caller.  */
        return (long)((unsigned long)(par_pid << 16) + (unsigned long)i);

}


/*
 *  This routine returns the clock freq. in Mhz of the local node.
 */
asmlinkage long
sys_gamma_init_part3
(void)
{

	int found=0,i=0,par_pid;
	par_entry *p;
	instance_entry *inst;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		p = par_proc_table[par_pid];
		if (p == NULL) continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			inst = &par_proc_table[par_pid]->inst_table[i];
			if (inst->pid == current->pid) {
				found = 1;
				break;
			}
		}
		if (found) break;
	}

	if (!found) return -GAMMA_ERR_GENERIC;

/*  Return the CPU clock freq. in MHz  */
        return cpu_khz/1000;

}


asmlinkage long
sys_gamma_exit
(u_int32_t par_pid, u_int32_t procnum)
{

	int i,j,found;
	par_entry *p;
	instance_entry *inst;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t *shared_ring;
#endif


/*  par_pid #0 is reserved, and par_pid in excess of max are forbidden  */
	if (par_pid == 0 || par_pid > MAX_PAR_PID)
		return -GAMMA_ERR_GENERIC;

	if (!spin_trylock(&gamma_rx_lock))
		return -GAMMA_ERR_TRYAGAIN;
	if (!spin_trylock(&gamma_tx_lock)) {
		spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!write_trylock(&gamma_rxintranode_lock)) {
		spin_unlock(&gamma_tx_lock);
		spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}
#endif

	p = par_proc_table[par_pid];

	if (p == NULL) {
/*  Virtual GAMMA #par_pid does not exist  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

	inst = &p->inst_table[procnum];

	if (inst->pid != current->pid) {
#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

/*  unsubscribe from virtual GAMMA  */
	inst->pid = -1;
	for (i=0; i<=MAX_PORT_NUMBER; i++)
		if (inst->port_table[i].header != NULL) {
			kfree(inst->port_table[i].header);
			inst->port_table[i].header = NULL;
		}

/*  if no other process instances left, destroy the virtual GAMMA  */
	found = 0;
	for (i=0; i<MAX_NUM_LOCAL_INST; i++)
		if (p->inst_table[i].pid > 0)
			found = 1;

	if (found) {
#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		spin_unlock(&gamma_rx_lock);
		return 0;
	}

	par_proc_table[par_pid] = NULL;
	gamma_must_yield.task = NULL;
	gamma_must_yield.pid = 0;
	gamma_must_yield.time.hi = 0;
	gamma_must_yield.time.lo = 0;
	gamma_run_handler.task = NULL;
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  Free all rings of temp buffers for intra-node communication.  */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		shared_ring = &p->inst_table[i].shared_ring;
		for (j=0; j < SHARED_RING_SIZE; j++)
			if (shared_ring->desc[j].buf != NULL) {
				kfree(shared_ring->desc[j].buf);
				shared_ring->desc[j].buf = NULL;
			}
	}
#endif
	free_pages((unsigned long)p,par_entry_magnitude);

#ifdef MORE_INSTANCES_ON_SAME_NODE
	write_unlock(&gamma_rxintranode_lock);
#endif
	spin_unlock(&gamma_tx_lock);
	spin_unlock(&gamma_rx_lock);

	return 0;

}


/*
 *  Destroy a virtual GAMMA and force all process instances to exit.
 *  There are two possible behaviours, namely:
 *
 *  global=1 : usually due to an error or signal (es. ctl-C) catched by one
 *  of the locally running process instances, this routine asks (by a SIGTERM)
 *  all other local instances to quit the specified GAMMA vm and propagate a
 *  similar request to all nodes in the network.  SIGTERM is catched and
 *  handled by this same routine.  On each node, the last process instance to
 *  quit the vm will also destroy the vm itself.
 *
 *  global=0 : usually due to a "gammareset" (invoked locally), this routine
 *  will kill (by SIGKILL) all local process instances on ==any== local GAMMA
 *  vm, then destroy all GAMMA vm's.  No similar request is propagated to
 *  other nodes in the network.
 *
 *  Flag "rxlock" is 0 if caller is gamma_kernel_rec() (so gamma_rx_lock is
 *  already taken), 1 otherwise (gamma_rx_lock still has to be taken).
 */
long
__sys_gamma_vm_abort
(u_int32_t key_or_parpid, int global, int rxlock)
{

	int i,j,found,par_pid;
	int was_not_already_quitting = 0;
	u_int32_t rnd_key;
	par_entry *p;
	instance_entry *inst;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t *shared_ring;
#endif


	if (rxlock)
		/* gamma_rx_lock not already locked, so take it as well */
		if (!spin_trylock(&gamma_rx_lock))
			return -GAMMA_ERR_TRYAGAIN;
	if (!spin_trylock(&gamma_tx_lock)) {
		if (rxlock) spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!write_trylock(&gamma_rxintranode_lock)) {
		spin_unlock(&gamma_tx_lock);
		if (rxlock) spin_unlock(&gamma_rx_lock);
		DELAY_FOR_SMP;
		return -GAMMA_ERR_TRYAGAIN;
	}
#endif

	if (global) {
/*  rnd_key of the virtual GAMMA is passed by parameter, find par_pid  */
		rnd_key = key_or_parpid;
		for (i=1; i<= MAX_PAR_PID; i++) {
			p = par_proc_table[i];
			if (p == NULL) continue;
			if (p->vm.random_key == rnd_key) break;
		}
		par_pid = i;
	} else {
/*  par_pid of the virtual GAMMA is passed by parameter  */
		par_pid = (int)key_or_parpid;
	}

	if (par_pid <= 0 || par_pid > MAX_PAR_PID) {
/*  Virtual GAMMA does not exist  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		if (rxlock) spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

	p = par_proc_table[par_pid];

	if (p == NULL) {
#ifdef MORE_INSTANCES_ON_SAME_NODE
		write_unlock(&gamma_rxintranode_lock);
#endif
		spin_unlock(&gamma_tx_lock);
		if (rxlock) spin_unlock(&gamma_rx_lock);
		return -GAMMA_ERR_GENERIC;
	}

	if (global) {

DEBUG(printg("sys_gamma_vm_abort: found par_pid %d...\n",par_pid));

/*  unsubscribe self from virtual GAMMA and dealloc own ports */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		inst = &p->inst_table[i];
		if (inst->pid != current->pid) continue;
		inst->pid = -1;
		for (j=0; j<=MAX_PORT_NUMBER; j++)
			if (inst->port_table[j].header != NULL) {
				kfree(inst->port_table[j].header);
				inst->port_table[j].header = NULL;
			}
	}

	} else {

/*  dealloc all instances' ports */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
		inst = &p->inst_table[i];
		if (inst->whoami < 0) continue;
		for (j=0; j<=MAX_PORT_NUMBER; j++)
			if (inst->port_table[j].header) {
				kfree(inst->port_table[j].header);
				inst->port_table[j].header = NULL;
			}
	}

	}

	if (global) {
/*
 *  Destroy the virtual GAMMA across the network and
 *  require remote process instances to quit (if not already done)
 */
		if (!p->quitting) {
			p->quitting = 1;
			was_not_already_quitting = 1;
/*  Some switches (3COM?) may need this, as they tend to sleep sometimes...  */
#if 0
			for (j=3;j;j--)
#endif
				gamma_kernel_send(BROADCAST,par_pid,VM_ABORT,
					&p->vm.random_key,
					sizeof(p->vm.random_key),0);
		}
	}
/*  Quit local process instances, if any  */
	for (i=0; i<MAX_NUM_LOCAL_INST; i++)
		if (
			p->inst_table[i].whoami >= 0 &&
			p->inst_table[i].task != NULL &&
			p->inst_table[i].task->state != EXIT_ZOMBIE &&
			p->inst_table[i].task->state != EXIT_DEAD &&
			p->inst_table[i].task->pid == p->inst_table[i].pid &&
			find_task_by_pid(p->inst_table[i].pid) != NULL
		) {
			if (global) {
				if (was_not_already_quitting)
		 		if (p->inst_table[i].task->pid
					!= current->pid) {

DEBUG_L(printg("sys_gamma_vm_abort: SIGTERM to inst %d\n",i));

					force_sig(SIGTERM,
						p->inst_table[i].task);
				}
			}
			else {

DEBUG_L(printg("sys_gamma_vm_abort: SIGKILL to inst %d\n",i));

				force_sig(SIGKILL, p->inst_table[i].task);
			}
		}

	gamma_must_yield.task = NULL;
	gamma_must_yield.pid = 0;
	gamma_must_yield.time.hi = 0;
	gamma_must_yield.time.lo = 0;
	gamma_run_handler.task = NULL;

/*  destroy the virtual GAMMA if no other instances still subscribed  */
	found = 0;
	if (global)
		for (i=0; i<MAX_NUM_LOCAL_INST; i++)
			if (p->inst_table[i].pid > 0)
				found = 1;
	if (!found) {
/*  either last instance, or not a global invocation; destroy vm anyway  */
		par_proc_table[par_pid] = NULL;
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  Free all rings of temp buffers for intra-node communication.  */
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			shared_ring = &p->inst_table[i].shared_ring;
			for (j=0; j < SHARED_RING_SIZE; j++)
				if (shared_ring->desc[j].buf != NULL) {
					kfree(shared_ring->desc[j].buf);
					shared_ring->desc[j].buf = NULL;
				}
		}
#endif
		free_pages((unsigned long)p,par_entry_magnitude);
	}

DEBUG_L(printg("sys_gamma_vm_abort: par_pid %u aborted\n",par_pid));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	write_unlock(&gamma_rxintranode_lock);
#endif
	spin_unlock(&gamma_tx_lock);
	if (rxlock) spin_unlock(&gamma_rx_lock);

	return 0;

}


asmlinkage long
sys_gamma_vm_abort
(u_int32_t key_or_parpid, int global)
{
	return __sys_gamma_vm_abort(key_or_parpid,global,1);
}


asmlinkage long
sys_gamma_getkey
(void)
{
	par_entry *p;
	int found=0,i=0,par_pid;

	for (par_pid=1; par_pid<=MAX_PAR_PID; par_pid++) {
		p = par_proc_table[par_pid];
		if (p == NULL) continue;
		for (i=0; i<MAX_NUM_LOCAL_INST; i++) {
			if (p->inst_table[i].pid != current->pid) continue;
			found = 1;
			break;
		}
		if (found) break;
	}
	if (found) {
/*  Return the key of the virtual GAMMA this process is registered with.  */
	        return p->vm.random_key;
	}
	return -GAMMA_ERR_GENERIC;
}
